import pytest
import json
import tempfile
from pathlib import Path
from unittest.mock import Mock, patch
from collections import defaultdict
from skylos.visitors.test_aware import TestAwareVisitor
from skylos.visitors.framework_aware import FrameworkAwareVisitor
from skylos.penalties import apply_penalties

from skylos.analyzer import Skylos, proc_file, analyze


@pytest.fixture
def mock_definition():
    def _create_mock_def(
        name,
        simple_name,
        type,
        references=0,
        is_exported=False,
        confidence=100,
        in_init=False,
        line=1,
    ):
        mock = Mock()
        mock.name = name
        mock.simple_name = simple_name
        mock.type = type
        mock.references = references
        mock.is_exported = is_exported
        mock.confidence = confidence
        mock.in_init = in_init
        mock.line = line
        mock.filename = Path("test.py")
        mock.skip_reason = None
        mock.why_confidence_reduced = []
        mock.conditional_import = False
        mock.to_dict.return_value = {
            "name": name,
            "type": type,
            "file": "test.py",
            "line": line,
        }
        return mock

    return _create_mock_def


@pytest.fixture
def mock_test_aware_visitor():
    mock = Mock(spec=TestAwareVisitor)
    mock.is_test_file = False
    mock.test_decorated_lines = set()
    return mock


@pytest.fixture
def mock_framework_aware_visitor():
    mock = Mock(spec=FrameworkAwareVisitor)
    mock.framework_decorated_lines = set()
    return mock


@pytest.fixture
def temp_python_project():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)

        main_py = temp_path / "main.py"
        main_py.write_text("""
def used_function():
    return "used"

def unused_function():
    return "unused"

class UsedClass:
    def method(self):
        pass

class UnusedClass:
    def method(self):
        pass

result = used_function()
instance = UsedClass()
""")

        package_dir = temp_path / "mypackage"
        package_dir.mkdir()

        init_py = package_dir / "__init__.py"
        init_py.write_text("""
from .module import exported_function

def internal_function():
    pass
""")

        module_py = package_dir / "module.py"
        module_py.write_text("""
def exported_function():
    return "exported"

def internal_function():
    return "internal"
""")

        yield temp_path


class TestSkylos:
    @pytest.fixture
    def skylos(self):
        return Skylos()

    def test_init(self, skylos):
        assert skylos.defs == {}
        assert skylos.refs == []
        assert skylos.dynamic == set()
        assert isinstance(skylos.exports, defaultdict)

    def test_module_name_generation(self, skylos):
        root = Path("/project")

        file_path = Path("/project/src/module.py")
        result = skylos._module(root, file_path)
        assert result == "module"

        file_path = Path("/project/src/__init__.py")
        result = skylos._module(root, file_path)
        assert result == ""

        file_path = Path("/project/src/package/submodule.py")
        result = skylos._module(root, file_path)
        assert result == "package.submodule"

        file_path = Path("/project/main.py")
        result = skylos._module(root, file_path)
        assert result == "main"

    def test_should_exclude_file(self, skylos):
        """
        should exclude pycache, build, egg-info and whatever is in exclude_folders
        """
        root = Path("/project")
        exclude_folders = {"__pycache__", "build", "*.egg-info"}

        file_path = Path("/project/src/__pycache__/module.pyc")
        assert skylos._should_exclude_file(file_path, root, exclude_folders)

        file_path = Path("/project/build/lib/module.py")
        assert skylos._should_exclude_file(file_path, root, exclude_folders)

        file_path = Path("/project/mypackage.egg-info/PKG-INFO")
        assert skylos._should_exclude_file(file_path, root, exclude_folders)

        file_path = Path("/project/src/module.py")
        assert not skylos._should_exclude_file(file_path, root, exclude_folders)

        assert not skylos._should_exclude_file(file_path, root, None)

    @patch("skylos.analyzer.Path")
    def test_get_python_files_single_file(self, mock_path, skylos):
        mock_file = Mock()
        mock_file.is_file.return_value = True
        mock_file.parent = Path("/project")
        mock_path.return_value.resolve.return_value = mock_file

        files, root = skylos._get_python_files("/project/test.py")
        assert files == [mock_file]
        assert root == Path("/project")

    @patch("skylos.analyzer.Path")
    def test_get_python_files_directory(self, mock_path, skylos):
        mock_dir = Mock()
        mock_dir.is_file.return_value = False
        mock_files = [Path("/project/file1.py"), Path("/project/file2.py")]

        mock_dir.glob.return_value = mock_files
        mock_path.return_value.resolve.return_value = mock_dir

        files, root = skylos._get_python_files("/project")

        assert mock_dir.glob.call_count == 4
        assert root == mock_dir

    def test_mark_exports_in_init(self, skylos):
        mock_def1 = Mock()
        mock_def1.in_init = True
        mock_def1.simple_name = "public_function"
        mock_def1.is_exported = False

        mock_def2 = Mock()
        mock_def2.in_init = True
        mock_def2.simple_name = "_private_function"
        mock_def2.is_exported = False

        skylos.defs = {
            "module.public_function": mock_def1,
            "module._private_function": mock_def2,
        }

        skylos._mark_exports()

        assert mock_def1.is_exported == True
        assert mock_def2.is_exported == False

    def test_mark_exports_explicit_exports(self, skylos):
        mock_def = Mock()
        mock_def.simple_name = "my_function"
        mock_def.type = "function"
        mock_def.is_exported = False
        mock_def.references = 0

        skylos.defs = {"module.my_function": mock_def}
        skylos.exports = {"module": {"my_function"}}

        skylos._mark_exports()

        assert mock_def.is_exported == True

    def test_mark_refs_direct_reference(self, skylos):
        mock_def = Mock()
        mock_def.type = "function"
        mock_def.simple_name = "function"
        mock_def.name = "module.function"
        mock_def.references = 0

        skylos.defs = {"module.function": mock_def}
        skylos.refs = [("module.function", None)]

        skylos._mark_refs()

        assert mock_def.references == 1

    def test_mark_refs_import_reference(self, skylos):
        mock_import = Mock()
        mock_import.type = "import"
        mock_import.simple_name = "imported_func"
        mock_import.name = "other_module.imported_func"
        mock_import.references = 0

        mock_original = Mock()
        mock_original.type = "function"
        mock_original.simple_name = "imported_func"
        mock_original.references = 0

        skylos.defs = {
            "module.imported_func": mock_import,
            "other_module.imported_func": mock_original,
        }
        skylos.refs = [("module.imported_func", None)]

        skylos._mark_refs()

        assert mock_import.references == 1
        assert mock_original.references == 2


class TestHeuristics:
    @pytest.fixture
    def skylos_with_class_methods(self, mock_definition):
        skylos = Skylos()

        mock_class = mock_definition(
            name="MyClass", simple_name="MyClass", type="class", references=1
        )

        mock_init = mock_definition(
            name="MyClass.__init__", simple_name="__init__", type="method", references=0
        )

        mock_enter = mock_definition(
            name="MyClass.__enter__",
            simple_name="__enter__",
            type="method",
            references=0,
        )

        skylos.defs = {
            "MyClass": mock_class,
            "MyClass.__init__": mock_init,
            "MyClass.__enter__": mock_enter,
        }

        return skylos, mock_class, mock_init, mock_enter

    def test_auto_called_methods_get_references(self, skylos_with_class_methods):
        """auto-called methods get reference counts when class is used."""
        skylos, _, mock_init, mock_enter = skylos_with_class_methods

        skylos._apply_heuristics()

        assert mock_init.references == 1
        assert mock_enter.references == 1


class TestAnalyze:
    @patch("skylos.analyzer.proc_file")
    def test_analyze_basic(self, mock_proc_file, temp_python_project):
        mock_def = Mock()
        mock_def.name = "test.unused_function"
        mock_def.references = 0
        mock_def.is_exported = False
        mock_def.confidence = 80
        mock_def.type = "function"
        mock_def.to_dict.return_value = {
            "name": "test.unused_function",
            "type": "function",
            "file": "test.py",
            "line": 1,
        }

        mock_def.line = 1
        mock_def.filename = "test.py"
        mock_def.simple_name = "unused_function"
        mock_def.in_init = False
        mock_def.skip_reason = None
        mock_def.filename = Path("test.py")

        mock_test_visitor = Mock(spec=TestAwareVisitor)
        mock_test_visitor.is_test_file = False
        mock_test_visitor.test_decorated_lines = set()

        mock_framework_visitor = Mock(spec=FrameworkAwareVisitor)
        mock_framework_visitor.framework_decorated_lines = set()
        mock_framework_visitor.is_framework_file = False

        mock_proc_file.return_value = (
            [mock_def],
            [],
            set(),
            set(),
            mock_test_visitor,
            mock_framework_visitor,
            [],
            [],
            [],
            None,
            None,
            None,
        )

        result_json = analyze(str(temp_python_project), conf=60)
        result = json.loads(result_json)

        assert "unused_functions" in result
        assert "unused_imports" in result
        assert "unused_classes" in result
        assert "unused_variables" in result
        assert "unused_parameters" in result
        assert "unused_files" in result
        assert "analysis_summary" in result

    def test_analyze_with_exclusions(self, temp_python_project):
        """analyze with folder exclusions."""
        exclude_dir = temp_python_project / "build"
        exclude_dir.mkdir()
        exclude_file = exclude_dir / "generated.py"
        exclude_file.write_text("def generated_function(): pass")

        result_json = analyze(str(temp_python_project), exclude_folders=["build"])
        result = json.loads(result_json)

        assert result["analysis_summary"]["excluded_folders"] == ["build"]

    def test_analyze_empty_directory(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            result_json = analyze(temp_dir, conf=60)
            result = json.loads(result_json)

            assert result["analysis_summary"]["total_files"] == 0
            assert all(
                len(result[key]) == 0
                for key in [
                    "unused_functions",
                    "unused_imports",
                    "unused_classes",
                    "unused_variables",
                    "unused_parameters",
                ]
            )

    def test_confidence_threshold_filtering(self, mock_definition):
        """confidence threshold properly filters results."""
        skylos = Skylos()

        high_conf = mock_definition(
            name="high_conf",
            simple_name="high_conf",
            type="function",
            references=0,
            is_exported=False,
            confidence=80,
        )

        low_conf = mock_definition(
            name="low_conf",
            simple_name="low_conf",
            type="function",
            references=0,
            is_exported=False,
            confidence=40,
        )

        skylos.defs = {"high_conf": high_conf, "low_conf": low_conf}

        with patch.object(skylos, "_get_python_files") as mock_get_files:
            mock_get_files.return_value = ([Path("/fake/file.py")], Path("/"))

            with patch("skylos.analyzer.proc_file") as mock_proc_file:
                mock_proc_file.return_value = (
                    [],
                    [],
                    set(),
                    set(),
                    Mock(spec=TestAwareVisitor),
                    Mock(spec=FrameworkAwareVisitor),
                    [],
                    [],
                    [],
                    None,
                    None,
                    None,
                )

                result_json = skylos.analyze("/fake/path", thr=60)
                result = json.loads(result_json)

                assert len(result["unused_functions"]) == 1
                assert result["unused_functions"][0]["name"] == "high_conf"


class TestProcFile:
    def test_proc_file_with_valid_python(self):
        with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False) as f:
            f.write("""
def test_function():
    return 42

class TestClass:
    def method(self):
        return "ok"
""")
            f.flush()

            try:
                with (
                    patch("skylos.analyzer.Visitor") as mock_visitor_class,
                    patch(
                        "skylos.analyzer.TestAwareVisitor"
                    ) as mock_test_visitor_class,
                    patch(
                        "skylos.analyzer.FrameworkAwareVisitor"
                    ) as mock_framework_visitor_class,
                ):
                    mock_visitor = Mock()
                    mock_visitor.defs = []
                    mock_visitor.refs = []
                    mock_visitor.dyn = set()
                    mock_visitor.exports = set()
                    mock_visitor.pattern_tracker = None
                    mock_visitor_class.return_value = mock_visitor

                    mock_test_visitor = Mock(spec=TestAwareVisitor)
                    mock_test_visitor_class.return_value = mock_test_visitor

                    mock_framework_visitor = Mock(spec=FrameworkAwareVisitor)
                    mock_framework_visitor_class.return_value = mock_framework_visitor

                    (
                        defs,
                        refs,
                        dyn,
                        exports,
                        test_flags,
                        framework_flags,
                        quality_findings,
                        danger_findings,
                        pro_findings,
                        pattern_tracker,
                        empty_file_finding,
                        cfg,
                        raw_imports,
                        ignore_lines,
                        suppressed_findings,
                        inferred_types,
                        instance_attr_types,
                        used_attr_names,
                        used_attr_context,
                        source_lines,
                    ) = proc_file(f.name, "test_module")

                    mock_visitor_class.assert_called_once_with("test_module", f.name)
                    mock_visitor.visit.assert_called_once()

                    assert defs == []
                    assert refs == []
                    assert dyn == set()
                    assert exports == set()
                    assert test_flags == mock_test_visitor
                    assert framework_flags == mock_framework_visitor
                    assert quality_findings == []
                    assert danger_findings == []
                    assert pro_findings == []
                    assert pattern_tracker is None
                    assert empty_file_finding is None
            finally:
                Path(f.name).unlink()

    def test_proc_file_with_invalid_python(self):
        with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False) as f:
            f.write("def invalid_syntax(:\npass")
            f.flush()

            try:
                (
                    defs,
                    refs,
                    dyn,
                    exports,
                    test_flags,
                    framework_flags,
                    quality_findings,
                    danger_findings,
                    pro_findings,
                    pattern_tracker,
                    empty_file_finding,
                    cfg,
                    raw_imports,
                    ignore_lines,
                    suppressed_findings,
                    inferred_types,
                    instance_attr_types,
                    used_attr_names,
                    used_attr_context,
                    source_lines,
                ) = proc_file(f.name, "test_module")

                assert defs == []
                assert refs == []
                assert dyn == set()
                assert exports == set()
                assert isinstance(test_flags, TestAwareVisitor)
                assert isinstance(framework_flags, FrameworkAwareVisitor)
                assert quality_findings == []
                assert danger_findings == []
                assert pro_findings == []
                assert pattern_tracker is None
                assert empty_file_finding is None
                assert isinstance(test_flags, TestAwareVisitor)
                assert isinstance(framework_flags, FrameworkAwareVisitor)
            finally:
                Path(f.name).unlink()

    def test_proc_file_with_tuple_args(self):
        with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False) as f:
            f.write("def test(): pass")
            f.flush()

            try:
                with (
                    patch("skylos.analyzer.Visitor") as mock_visitor_class,
                    patch(
                        "skylos.analyzer.TestAwareVisitor"
                    ) as mock_test_visitor_class,
                    patch(
                        "skylos.analyzer.FrameworkAwareVisitor"
                    ) as mock_framework_visitor_class,
                ):
                    mock_visitor = Mock()
                    mock_visitor.defs = []
                    mock_visitor.refs = []
                    mock_visitor.dyn = set()
                    mock_visitor.exports = set()
                    mock_visitor.pattern_tracker = None
                    mock_visitor_class.return_value = mock_visitor

                    mock_test_visitor = Mock(spec=TestAwareVisitor)
                    mock_test_visitor_class.return_value = mock_test_visitor

                    mock_framework_visitor = Mock(spec=FrameworkAwareVisitor)
                    mock_framework_visitor_class.return_value = mock_framework_visitor

                    (
                        defs,
                        refs,
                        dyn,
                        exports,
                        test_flags,
                        framework_flags,
                        quality_findings,
                        danger_findings,
                        pro_findings,
                        pattern_tracker,
                        empty_file_finding,
                        cfg,
                        raw_imports,
                        ignore_lines,
                        suppressed_findings,
                        inferred_types,
                        instance_attr_types,
                        used_attr_names,
                        used_attr_context,
                        source_lines,
                    ) = proc_file((f.name, "test_module"))

                    mock_visitor_class.assert_called_once_with("test_module", f.name)
            finally:
                Path(f.name).unlink()

    def test_empty_file_reporting(self, tmp_path):
        empty = tmp_path / "empty_module.py"
        empty.write_text("")

        (tmp_path / "main.py").write_text("")
        pkg = tmp_path / "mypkg"
        pkg.mkdir()
        (pkg / "__init__.py").write_text('"""package init docstring"""')

        result_json = analyze(str(tmp_path), conf=0)
        result = json.loads(result_json)

        assert "unused_files" in result
        files = result["unused_files"]

        flagged = {Path(f["file"]).name for f in files}
        assert "empty_module.py" in flagged
        assert "main.py" not in flagged
        assert "__init__.py" not in flagged

        item = next(f for f in files if Path(f["file"]).name == "empty_module.py")
        assert item["rule_id"] == "SKY-E002"
        assert item["category"] == "DEAD_CODE"
        assert item["severity"] == "LOW"


class TestApplyPenalties:
    @pytest.mark.parametrize(
        ("filename", "def_type", "expected_reason"),
        [
            ("tests/test_api.py", "function", "test-only path"),
            ("examples/demo.py", "function", "standalone example path"),
            ("benchmarks/bench_api.py", "class", "benchmark entrypoint path"),
        ],
    )
    @patch("skylos.penalties.detect_framework_usage")
    def test_non_library_paths_suppress_dead_code_callables(
        self,
        mock_detect_framework,
        filename,
        def_type,
        expected_reason,
        mock_definition,
        mock_test_aware_visitor,
        mock_framework_aware_visitor,
    ):
        mock_detect_framework.return_value = None

        skylos = Skylos()
        mock_def = mock_definition(
            name="demo.symbol",
            simple_name="symbol",
            type=def_type,
            confidence=100,
        )
        mock_def.filename = Path(filename)

        apply_penalties(
            skylos, mock_def, mock_test_aware_visitor, mock_framework_aware_visitor
        )

        assert mock_def.confidence == 0
        assert mock_def.skip_reason == expected_reason

    @patch("skylos.penalties.detect_framework_usage")
    def test_private_name_penalty(
        self,
        mock_detect_framework,
        mock_definition,
        mock_test_aware_visitor,
        mock_framework_aware_visitor,
    ):
        mock_detect_framework.return_value = None

        skylos = Skylos()
        mock_def = mock_definition(
            name="_private_func",
            simple_name="_private_func",
            type="function",
            confidence=100,
        )

        apply_penalties(
            skylos, mock_def, mock_test_aware_visitor, mock_framework_aware_visitor
        )
        assert mock_def.confidence < 100

    @patch("skylos.penalties.detect_framework_usage")
    def test_magic_methods_confidence_zero(
        self,
        mock_detect_framework,
        mock_definition,
        mock_test_aware_visitor,
        mock_framework_aware_visitor,
    ):
        """magic methods get confidence of 0."""
        mock_detect_framework.return_value = None
        skylos = Skylos()
        mock_def = mock_definition(
            name="MyClass.__str__", simple_name="__str__", type="method", confidence=100
        )

        apply_penalties(
            skylos, mock_def, mock_test_aware_visitor, mock_framework_aware_visitor
        )
        assert mock_def.confidence == 0

    @patch("skylos.penalties.detect_framework_usage")
    def test_self_cls_parameters_confidence_zero(
        self,
        mock_detect_framework,
        mock_definition,
        mock_test_aware_visitor,
        mock_framework_aware_visitor,
    ):
        mock_detect_framework.return_value = None
        skylos = Skylos()

        mock_self = mock_definition(
            name="MyClass.method.self",
            simple_name="self",
            type="parameter",
            confidence=100,
        )

        mock_cls = mock_definition(
            name="MyClass.classmethod.cls",
            simple_name="cls",
            type="parameter",
            confidence=100,
        )

        apply_penalties(
            skylos, mock_self, mock_test_aware_visitor, mock_framework_aware_visitor
        )
        apply_penalties(
            skylos, mock_cls, mock_test_aware_visitor, mock_framework_aware_visitor
        )

        assert mock_self.confidence == 0
        assert mock_cls.confidence == 0

    @patch("skylos.penalties.detect_framework_usage")
    def test_conditional_import_penalty_reduces_confidence(
        self,
        mock_detect_framework,
        mock_definition,
        mock_test_aware_visitor,
        mock_framework_aware_visitor,
    ):
        mock_detect_framework.return_value = None
        skylos = Skylos()

        mock_def = mock_definition(
            name="brotli",
            simple_name="brotli",
            type="import",
            confidence=100,
        )
        mock_def.conditional_import = True

        apply_penalties(
            skylos, mock_def, mock_test_aware_visitor, mock_framework_aware_visitor
        )

        assert mock_def.confidence == 40
        assert "conditional_import_fallback" in mock_def.why_confidence_reduced

    @patch("skylos.penalties.detect_framework_usage")
    def test_test_methods_confidence_zero(
        self, mock_detect_framework, mock_definition, mock_framework_aware_visitor
    ):
        """test methods get confidence of 0"""
        mock_detect_framework.return_value = None

        skylos = Skylos()

        mock_def = mock_definition(
            name="TestMyClass.test_something",
            simple_name="test_something",
            type="method",
            confidence=100,
        )

        test_visitor = Mock(spec=TestAwareVisitor)
        test_visitor.is_test_file = True
        test_visitor.test_decorated_lines = {mock_def.line}

        apply_penalties(skylos, mock_def, test_visitor, mock_framework_aware_visitor)
        assert mock_def.confidence == 0

    @patch("skylos.penalties.detect_framework_usage")
    def test_underscore_variable_confidence_zero(
        self,
        mock_detect_framework,
        mock_definition,
        mock_test_aware_visitor,
        mock_framework_aware_visitor,
    ):
        """underscore variables get confidence of 0."""
        mock_detect_framework.return_value = None

        skylos = Skylos()

        mock_def = mock_definition(
            name="_", simple_name="_", type="variable", confidence=100
        )

        apply_penalties(
            skylos, mock_def, mock_test_aware_visitor, mock_framework_aware_visitor
        )
        assert mock_def.confidence == 0


class TestIgnorePragmas:
    def test_analyze_respects_ignore_pragmas(self, tmp_path):
        src = tmp_path / "demo.py"
        src.write_text(
            """
def used():
    pass

def unused_no_ignore():
    pass

def unused_ignore():   # pragma: no skylos
    pass

used()
"""
        )

        result_json = analyze(str(tmp_path), conf=0)
        result = json.loads(result_json)

        unreachable = {
            item["name"].split(".")[-1] for item in result["unused_functions"]
        }

        assert "unused_no_ignore" in unreachable
        assert "unused_ignore" not in unreachable
        assert "used" not in unreachable


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



import logging
import os
import re
import subprocess
from dataclasses import dataclass, field
from pathlib import Path

from skylos.constants import NETWORK_TIMEOUT_SHORT, SUBPROCESS_TIMEOUT

logger = logging.getLogger(__name__)

AI_COAUTHOR_PATTERNS = [
    re.compile(r"copilot", re.IGNORECASE),
    re.compile(r"claude", re.IGNORECASE),
    re.compile(r"cursor", re.IGNORECASE),
    re.compile(r"codewhisperer", re.IGNORECASE),
    re.compile(r"tabnine", re.IGNORECASE),
    re.compile(r"github-actions\[bot\]", re.IGNORECASE),
    re.compile(r"devin", re.IGNORECASE),
    re.compile(r"codex", re.IGNORECASE),
    re.compile(r"aider", re.IGNORECASE),
]

AI_EMAIL_PATTERNS = [
    re.compile(r"\[bot\]@", re.IGNORECASE),
    re.compile(r"copilot", re.IGNORECASE),
    re.compile(r"cursor", re.IGNORECASE),
    re.compile(r"claude", re.IGNORECASE),
    re.compile(r"noreply@anthropic\.com", re.IGNORECASE),
    re.compile(r"noreply@github\.com", re.IGNORECASE),
]

AI_MESSAGE_PATTERNS = [
    re.compile(r"generated\s+by\s+(copilot|claude|cursor|ai|codex)", re.IGNORECASE),
    re.compile(r"ai[- ]generated", re.IGNORECASE),
    re.compile(r"co-authored-by.*copilot", re.IGNORECASE),
    re.compile(r"co-authored-by.*claude", re.IGNORECASE),
    re.compile(r"co-authored-by.*cursor", re.IGNORECASE),
    re.compile(r"co-authored-by.*devin", re.IGNORECASE),
]

AGENT_NAME_MAP = {
    "copilot": "copilot",
    "claude": "claude",
    "cursor": "cursor",
    "codewhisperer": "codewhisperer",
    "tabnine": "tabnine",
    "devin": "devin",
    "codex": "codex",
    "aider": "aider",
    "anthropic": "claude",
    "github-actions[bot]": "github-actions",
}

HUNK_HEADER_RE = re.compile(r"^@@ -\d+(?:,\d+)? \+(\d+)(?:,(\d+))? @@")


@dataclass
class FileProvenance:
    file_path: str
    agent_authored: bool = False
    agent_lines: list = field(default_factory=list)
    indicators: list = field(default_factory=list)
    agent_name: str | None = None


@dataclass
class ProvenanceReport:
    files: dict = field(default_factory=dict)
    agent_files: list = field(default_factory=list)
    human_files: list = field(default_factory=list)
    summary: dict = field(default_factory=dict)
    confidence: str = "low"

    def to_dict(self):
        file_entries = {}
        for path, fp in self.files.items():
            file_entries[path] = {
                "file_path": fp.file_path,
                "agent_authored": fp.agent_authored,
                "agent_lines": fp.agent_lines,
                "indicators": fp.indicators,
                "agent_name": fp.agent_name,
            }
        return {
            "files": file_entries,
            "agent_files": self.agent_files,
            "human_files": self.human_files,
            "summary": self.summary,
            "confidence": self.confidence,
        }


def _detect_agent_name(text):
    text_lower = text.lower()
    for keyword, name in AGENT_NAME_MAP.items():
        if keyword in text_lower:
            return name
    return None


def _parse_diff_hunks(diff_text):
    file_ranges = {}
    current_file = None

    for line in diff_text.splitlines():
        if line.startswith("+++ b/"):
            current_file = line[6:]
        elif line.startswith("+++ /dev/null"):
            current_file = None
        elif line.startswith("@@") and current_file is not None:
            m = HUNK_HEADER_RE.match(line)
            if m:
                start = int(m.group(1))
                count = int(m.group(2)) if m.group(2) else 1
                end = start + max(count - 1, 0)
                if current_file not in file_ranges:
                    file_ranges[current_file] = []
                file_ranges[current_file].append((start, end))

    return file_ranges


def _resolve_base_ref(explicit_base=None):
    if explicit_base:
        return explicit_base

    env_base = os.environ.get("GITHUB_BASE_REF")
    if env_base:
        return f"origin/{env_base}"

    return "origin/main"


def _git_merge_base(git_root, base_ref):
    try:
        return (
            subprocess.check_output(
                ["git", "merge-base", base_ref, "HEAD"],
                cwd=git_root,
                stderr=subprocess.DEVNULL,
                timeout=NETWORK_TIMEOUT_SHORT,
            )
            .decode("utf-8", errors="ignore")
            .strip()
        )
    except (subprocess.SubprocessError, OSError):
        return None


def analyze_provenance(git_root, base_ref=None):
    if not git_root:
        return ProvenanceReport()

    base_ref = _resolve_base_ref(base_ref)
    merge_base = _git_merge_base(git_root, base_ref)

    if not merge_base:
        logger.debug(
            "Could not find merge base for %s, falling back to HEAD~10", base_ref
        )
        range_spec = "HEAD~10..HEAD"
    else:
        range_spec = f"{merge_base}..HEAD"

    indicators_by_commit = {}
    ai_commits = set()
    agents_seen = set()

    try:
        log_output = subprocess.check_output(
            [
                "git",
                "log",
                "--format=%H|%an|%ae|%s|%(trailers:key=Co-authored-by,valueonly,separator=%x00)",
                range_spec,
            ],
            cwd=git_root,
            stderr=subprocess.DEVNULL,
            timeout=SUBPROCESS_TIMEOUT,
        ).decode("utf-8", errors="ignore")
    except (subprocess.SubprocessError, OSError):
        logger.debug("Failed to get git log for provenance", exc_info=True)
        return ProvenanceReport()

    for line in log_output.strip().splitlines():
        if not line.strip():
            continue
        parts = line.split("|", 4)
        if len(parts) < 4:
            continue

        commit_sha = parts[0]
        author_name = parts[1]
        author_email = parts[2]
        subject = parts[3]
        trailers = parts[4] if len(parts) > 4 else ""

        is_ai_commit = False
        indicator = None

        for pat in AI_COAUTHOR_PATTERNS:
            if pat.search(trailers):
                agent = _detect_agent_name(trailers)
                indicator = {
                    "type": "co-author",
                    "commit": commit_sha[:7],
                    "detail": trailers.strip()[:100],
                    "agent_name": agent,
                }
                is_ai_commit = True
                if agent:
                    agents_seen.add(agent)
                break

        if not is_ai_commit:
            for pat in AI_EMAIL_PATTERNS:
                if pat.search(author_email):
                    agent = _detect_agent_name(author_email) or _detect_agent_name(
                        author_name
                    )
                    indicator = {
                        "type": "author-email",
                        "commit": commit_sha[:7],
                        "detail": f"{author_name} <{author_email}>",
                        "agent_name": agent,
                    }
                    is_ai_commit = True
                    if agent:
                        agents_seen.add(agent)
                    break

        if not is_ai_commit:
            for pat in AI_MESSAGE_PATTERNS:
                if pat.search(subject):
                    agent = _detect_agent_name(subject)
                    indicator = {
                        "type": "commit-message",
                        "commit": commit_sha[:7],
                        "detail": subject[:100],
                        "agent_name": agent,
                    }
                    is_ai_commit = True
                    if agent:
                        agents_seen.add(agent)
                    break

        if is_ai_commit:
            ai_commits.add(commit_sha)
            indicators_by_commit[commit_sha] = indicator

    file_provenance = {}
    all_changed_files = set()

    for commit_sha in ai_commits:
        try:
            diff_output = subprocess.check_output(
                ["git", "diff-tree", "-p", "-r", "--no-commit-id", commit_sha],
                cwd=git_root,
                stderr=subprocess.DEVNULL,
                timeout=SUBPROCESS_TIMEOUT,
            ).decode("utf-8", errors="ignore")
        except (subprocess.SubprocessError, OSError):
            logger.debug(
                "Failed to get diff-tree for %s", commit_sha[:7], exc_info=True
            )
            continue

        file_ranges = _parse_diff_hunks(diff_output)
        indicator = indicators_by_commit.get(commit_sha, {})

        for fpath, ranges in file_ranges.items():
            all_changed_files.add(fpath)
            if fpath not in file_provenance:
                file_provenance[fpath] = FileProvenance(
                    file_path=fpath,
                    agent_authored=True,
                    agent_lines=[],
                    indicators=[],
                    agent_name=indicator.get("agent_name"),
                )
            fp = file_provenance[fpath]
            fp.agent_lines.extend(ranges)
            fp.indicators.append(indicator)
            if not fp.agent_name and indicator.get("agent_name"):
                fp.agent_name = indicator["agent_name"]

    for fp in file_provenance.values():
        fp.agent_lines = _merge_ranges(fp.agent_lines)

    try:
        all_files_output = subprocess.check_output(
            ["git", "diff", "--name-only", range_spec],
            cwd=git_root,
            stderr=subprocess.DEVNULL,
            timeout=SUBPROCESS_TIMEOUT,
        ).decode("utf-8", errors="ignore")
        all_pr_files = {
            f.strip() for f in all_files_output.strip().splitlines() if f.strip()
        }
    except (subprocess.SubprocessError, OSError):
        all_pr_files = all_changed_files

    agent_files = sorted(file_provenance.keys())
    human_files = sorted(all_pr_files - set(agent_files))

    for hf in human_files:
        file_provenance[hf] = FileProvenance(file_path=hf, agent_authored=False)

    total = len(all_pr_files)
    agent_count = len(agent_files)
    indicator_count = sum(len(fp.indicators) for fp in file_provenance.values())

    if indicator_count > 5:
        confidence = "high"
    elif indicator_count > 0:
        confidence = "medium"
    else:
        confidence = "low"

    return ProvenanceReport(
        files=file_provenance,
        agent_files=agent_files,
        human_files=human_files,
        summary={
            "total_files": total,
            "agent_count": agent_count,
            "human_count": len(human_files),
            "agents_seen": sorted(agents_seen),
        },
        confidence=confidence,
    )


def _merge_ranges(ranges):
    if not ranges:
        return []
    sorted_ranges = sorted(ranges, key=lambda r: r[0])
    merged = [sorted_ranges[0]]
    for start, end in sorted_ranges[1:]:
        prev_start, prev_end = merged[-1]
        if start <= prev_end + 1:
            merged[-1] = (prev_start, max(prev_end, end))
        else:
            merged.append((start, end))
    return merged


def _line_in_ranges(line, ranges):
    for start, end in ranges:
        if start <= line <= end:
            return True
    return False


def annotate_findings_with_provenance(
    findings: list[dict],
    provenance_report: "ProvenanceReport",
) -> list[dict]:
    for finding in findings:
        file_path = finding.get("file")
        if not file_path:
            finding["ai_authored"] = False
            finding["ai_agent"] = None
            continue

        file_prov = provenance_report.files.get(file_path)

        if file_prov is None:
            for prov_path, prov in provenance_report.files.items():
                if file_path.endswith(prov_path) or prov_path.endswith(file_path):
                    file_prov = prov
                    break

        if file_prov is None or not file_prov.agent_authored:
            finding["ai_authored"] = False
            finding["ai_agent"] = None
            continue

        line = finding.get("line")
        if file_prov.agent_lines and line is not None:
            if _line_in_ranges(line, file_prov.agent_lines):
                finding["ai_authored"] = True
                finding["ai_agent"] = file_prov.agent_name
            else:
                finding["ai_authored"] = False
                finding["ai_agent"] = None
        else:
            finding["ai_authored"] = True
            finding["ai_agent"] = file_prov.agent_name

    return findings


def compute_ai_security_stats(
    findings: list[dict],
) -> dict:
    total = len(findings)
    ai_count = sum(1 for f in findings if f.get("ai_authored"))
    pct = (ai_count / total * 100) if total > 0 else 0.0

    by_agent: dict[str, int] = {}
    by_severity: dict[str, dict[str, int]] = {}
    by_category: dict[str, dict[str, int]] = {}

    for f in findings:
        is_ai = bool(f.get("ai_authored"))
        agent = f.get("ai_agent")
        severity = (f.get("severity") or "UNKNOWN").upper()
        category = (f.get("category") or "unknown").lower()

        if is_ai and agent:
            by_agent[agent] = by_agent.get(agent, 0) + 1

        if severity not in by_severity:
            by_severity[severity] = {"total": 0, "ai": 0}
        by_severity[severity]["total"] += 1
        if is_ai:
            by_severity[severity]["ai"] += 1

        if category not in by_category:
            by_category[category] = {"total": 0, "ai": 0}
        by_category[category]["total"] += 1
        if is_ai:
            by_category[category]["ai"] += 1

    return {
        "total_findings": total,
        "ai_authored_findings": ai_count,
        "ai_authored_pct": round(pct, 1),
        "by_agent": by_agent,
        "by_severity": by_severity,
        "by_category": by_category,
    }


@dataclass
class RiskIntersection:
    high_risk: list = field(default_factory=list)
    medium_risk: list = field(default_factory=list)
    summary: dict = field(default_factory=dict)

    def to_dict(self):
        return {
            "high_risk": self.high_risk,
            "medium_risk": self.medium_risk,
            "summary": self.summary,
        }


def compute_risk_intersections(git_root, provenance_report, exclude_folders=None):
    from skylos.discover.detector import detect_integrations
    from skylos.defend.engine import run_defense_checks

    if not provenance_report.agent_files:
        return RiskIntersection(summary={"high": 0, "medium": 0, "total_ai_files": 0})

    scan_path = Path(git_root)
    integrations, graph = detect_integrations(
        scan_path, exclude_folders=exclude_folders
    )
    defense_results, defense_score, ops_score = run_defense_checks(integrations, graph)

    integration_files = set()
    for integration in integrations:
        loc = integration.location
        if ":" in loc:
            loc = loc.split(":")[0]
        try:
            rel = str(Path(loc).relative_to(scan_path))
        except ValueError:
            rel = loc
        integration_files.add(rel)

    failed_defense_files = set()
    for result in defense_results:
        if not result.passed:
            loc = result.location
            if ":" in loc:
                loc = loc.split(":")[0]
            try:
                rel = str(Path(loc).relative_to(scan_path))
            except ValueError:
                rel = loc
            failed_defense_files.add(rel)

    high_risk = []
    medium_risk = []

    for agent_file in provenance_report.agent_files:
        has_integration = agent_file in integration_files
        has_failed_defense = agent_file in failed_defense_files

        if has_integration and has_failed_defense:
            high_risk.append(
                {
                    "file_path": agent_file,
                    "agent_name": provenance_report.files[agent_file].agent_name,
                    "reasons": [
                        "ai_authored",
                        "has_llm_integration",
                        "failed_defense_check",
                    ],
                }
            )
        elif has_integration or has_failed_defense:
            reasons = ["ai_authored"]
            if has_integration:
                reasons.append("has_llm_integration")
            if has_failed_defense:
                reasons.append("failed_defense_check")
            medium_risk.append(
                {
                    "file_path": agent_file,
                    "agent_name": provenance_report.files[agent_file].agent_name,
                    "reasons": reasons,
                }
            )

    return RiskIntersection(
        high_risk=high_risk,
        medium_risk=medium_risk,
        summary={
            "high": len(high_risk),
            "medium": len(medium_risk),
            "total_ai_files": len(provenance_report.agent_files),
        },
    )



from __future__ import annotations

import base64
import re
import unicodedata


ZERO_WIDTH_CHARS: set[str] = {
    "\u200b",
    "\u200c",
    "\u200d",
    "\u2060",
    "\ufeff",
    "\u200e",
    "\u200f",
    "\u202a",
    "\u202b",
    "\u202c",
    "\u202d",
    "\u202e",
}

_zw_escaped = []
for c in ZERO_WIDTH_CHARS:
    _zw_escaped.append(re.escape(c))
_ZERO_WIDTH_RE = re.compile("[" + "".join(_zw_escaped) + "]")

# common cyrillic/greek characters that look like latin ASCII
_CONFUSABLES: dict[str, str] = {
    "\u0410": "A",  # Cyrillic А
    "\u0412": "B",  # Cyrillic В
    "\u0421": "C",  # Cyrillic С
    "\u0415": "E",  # Cyrillic Е
    "\u041d": "H",  # Cyrillic Н
    "\u041a": "K",  # Cyrillic К
    "\u041c": "M",  # Cyrillic М
    "\u041e": "O",  # Cyrillic О
    "\u0420": "P",  # Cyrillic Р
    "\u0422": "T",  # Cyrillic Т
    "\u0425": "X",  # Cyrillic Х
    "\u0430": "a",  # Cyrillic а
    "\u0435": "e",  # Cyrillic е
    "\u043e": "o",  # Cyrillic о
    "\u0440": "p",  # Cyrillic р
    "\u0441": "c",  # Cyrillic с
    "\u0443": "y",  # Cyrillic у
    "\u0445": "x",  # Cyrillic х
    "\u0455": "s",  # Cyrillic ѕ
    "\u0456": "i",  # Cyrillic і
    "\u0458": "j",  # Cyrillic ј
    "\u0391": "A",  # Greek Α
    "\u0392": "B",  # Greek Β
    "\u0395": "E",  # Greek Ε
    "\u0397": "H",  # Greek Η
    "\u0399": "I",  # Greek Ι
    "\u039a": "K",  # Greek Κ
    "\u039c": "M",  # Greek Μ
    "\u039d": "N",  # Greek Ν
    "\u039f": "O",  # Greek Ο
    "\u03a1": "P",  # Greek Ρ
    "\u03a4": "T",  # Greek Τ
    "\u03a7": "X",  # Greek Χ
    "\u03bf": "o",  # Greek ο
}

_conf_escaped = []
for c in _CONFUSABLES:
    _conf_escaped.append(re.escape(c))
_CONFUSABLE_RE = re.compile("[" + "".join(_conf_escaped) + "]")

_BASE64_TOKEN_RE = re.compile(r"[A-Za-z0-9+/]{16,}={0,2}")


def normalize(text: str) -> str:
    text = unicodedata.normalize("NFKC", text)
    text = re.sub(r"[ \t]+", " ", text)
    return text


def strip_zero_width(text: str) -> tuple[str, list[tuple[str, int]]]:
    found: list[tuple[str, int]] = []
    for match in _ZERO_WIDTH_RE.finditer(text):
        pos = match.start()
        line_no = text[:pos].count("\n") + 1
        char_hex = f"U+{ord(match.group()):04X}"
        found.append((char_hex, line_no))

    cleaned = _ZERO_WIDTH_RE.sub("", text)
    return cleaned, found


def decode_base64_blobs(text: str) -> list[tuple[str, int]]:
    results: list[tuple[str, int]] = []
    for match in _BASE64_TOKEN_RE.finditer(text):
        token = match.group()
        line_no = text[: match.start()].count("\n") + 1

        try:
            decoded_bytes = base64.b64decode(token, validate=True)
            decoded = decoded_bytes.decode("utf-8", errors="strict")
        except Exception:
            continue

        if (
            len(decoded) >= 8
            and " " in decoded
            and sum(1 for c in decoded if c.isprintable()) / len(decoded) > 0.8
        ):
            results.append((decoded, line_no))

    return results


def detect_homoglyphs(text: str) -> list[tuple[str, str, int]]:
    if not _CONFUSABLE_RE.search(text):
        return []

    results: list[tuple[str, str, int]] = []

    for line_no, line in enumerate(text.splitlines(), 1):
        for word_match in re.finditer(r"\S+", line):
            word = word_match.group()
            has_ascii = bool(re.search(r"[a-zA-Z]", word))
            has_confusable = bool(_CONFUSABLE_RE.search(word))
            if not (has_ascii and has_confusable):
                continue
            for char_match in _CONFUSABLE_RE.finditer(word):
                char = char_match.group()
                results.append((char, _CONFUSABLES[char], line_no))

    return results



#!/usr/bin/env python3
import pytest
import ast
from unittest.mock import Mock, patch

from skylos.visitors.framework_aware import (
    FrameworkAwareVisitor,
    detect_framework_usage,
    FRAMEWORK_DECORATORS,
    FRAMEWORK_FUNCTIONS,
    FRAMEWORK_IMPORTS,
)


class TestFrameworkAwareVisitor:
    def test_init_default(self):
        v = FrameworkAwareVisitor()
        assert v.is_framework_file is False
        assert v.framework_decorated_lines == set()
        assert v.detected_frameworks == set()

    def test_flask_import_detection(self):
        code = """
import flask
from flask import Flask, request
"""
        tree = ast.parse(code)
        v = FrameworkAwareVisitor()
        v.visit(tree)
        v.finalize()
        assert v.is_framework_file is True
        assert "flask" in v.detected_frameworks

    def test_fastapi_import_detection(self):
        code = """
from fastapi import FastAPI
import fastapi
"""
        tree = ast.parse(code)
        v = FrameworkAwareVisitor()
        v.visit(tree)
        v.finalize()
        assert v.is_framework_file is True
        assert "fastapi" in v.detected_frameworks

    def test_django_import_detection(self):
        code = """
from django.http import HttpResponse
from django.views import View
"""
        tree = ast.parse(code)
        v = FrameworkAwareVisitor()
        v.visit(tree)
        v.finalize()
        assert v.is_framework_file is True
        assert "django" in v.detected_frameworks

    def test_flask_route_decorator_detection(self):
        code = """
@app.route('/api/users')
def get_users():
    return []

@app.post('/api/users')
def create_user():
    return {}
"""
        tree = ast.parse(code)
        v = FrameworkAwareVisitor()
        v.visit(tree)
        v.finalize()
        assert v.is_framework_file is True
        assert 3 in v.framework_decorated_lines
        assert 7 in v.framework_decorated_lines

    def test_fastapi_router_decorator_detection(self):
        code = """
@router.get('/items')
async def read_items():
    return []

@router.post('/items')
async def create_item():
    return {}
"""
        tree = ast.parse(code)
        v = FrameworkAwareVisitor()
        v.visit(tree)
        v.finalize()
        assert v.is_framework_file is True
        assert 3 in v.framework_decorated_lines
        assert 7 in v.framework_decorated_lines

    def test_django_decorator_detection(self):
        code = """
@login_required
def protected_view(request):
    return HttpResponse("Protected")

@permission_required('auth.add_user')
def admin_view(request):
    return HttpResponse("Admin")
"""
        tree = ast.parse(code)
        v = FrameworkAwareVisitor()
        v.visit(tree)
        v.finalize()
        assert v.is_framework_file is True
        assert 3 in v.framework_decorated_lines
        assert 7 in v.framework_decorated_lines

    def test_django_view_class_detection(self):
        code = """
from django import views

class UserView(View):
    def get(self, request):
        return HttpResponse("GET")

class UserViewSet(ViewSet):
    def list(self, request):
        return Response([])
"""
        tree = ast.parse(code)
        v = FrameworkAwareVisitor()
        v.visit(tree)
        v.finalize()
        assert v.is_framework_file is True
        assert 5 in v.framework_decorated_lines
        assert 9 in v.framework_decorated_lines

    def test_framework_functions_not_detected_in_non_framework_file(self):
        code = """
def save(self):
    pass

def get(self):
    pass
"""
        tree = ast.parse(code)
        v = FrameworkAwareVisitor()
        v.visit(tree)
        v.finalize()
        assert v.is_framework_file is False
        assert v.framework_decorated_lines == set()

    def test_multiple_decorators(self):
        code = """
@app.route('/users')
@login_required
@cache.cached(timeout=60)
def get_users():
    return []
"""
        tree = ast.parse(code)
        v = FrameworkAwareVisitor()
        v.visit(tree)
        v.finalize()
        assert v.is_framework_file is True
        assert 5 in v.framework_decorated_lines

    def test_complex_decorator_patterns(self):
        code = """
@app.route('/api/v1/users/<int:user_id>', methods=['GET', 'POST'])
def user_endpoint(user_id):
    return {}

@router.get('/items/{item_id}')
async def get_item(item_id: int):
    return {}
"""
        tree = ast.parse(code)
        v = FrameworkAwareVisitor()
        v.visit(tree)
        v.finalize()
        assert v.is_framework_file is True
        assert 3 in v.framework_decorated_lines
        assert 7 in v.framework_decorated_lines

    @patch("skylos.visitors.framework_aware.Path")
    def test_file_content_framework_detection(self, mock_path):
        mock_file = Mock()
        mock_file.read_text.return_value = (
            "from flask import Flask\napp = Flask(__name__)"
        )
        mock_path.return_value = mock_file
        v = FrameworkAwareVisitor(filename="test.py")
        v.finalize()
        assert v.is_framework_file is True
        assert "flask" in v.detected_frameworks

    def test_normalize_decorator_name(self):
        v = FrameworkAwareVisitor()
        node = ast.parse("@decorator\ndef func(): pass").body[0].decorator_list[0]
        assert v._normalize_decorator(node) == "@decorator"
        node = ast.parse("@app.route\ndef func(): pass").body[0].decorator_list[0]
        assert v._normalize_decorator(node) == "@app.route"

    def test_depends_marks_dependency_and_flags_framework_file(self):
        code = """
from fastapi import Depends

def dep():
    return 1

@router.get("/")
def foo(x: int = Depends(dep)):
    return {}
"""
        tree = ast.parse(code)
        v = FrameworkAwareVisitor()
        v.visit(tree)
        v.finalize()
        assert v.is_framework_file is True
        assert 4 in v.framework_decorated_lines

    def test_typed_model_in_route_marks_model_definition(self):
        code = """
from pydantic import BaseModel

class In(BaseModel):
    x: int

@router.post("/")
def calc(req: In):
    return 1
"""
        tree = ast.parse(code)
        v = FrameworkAwareVisitor()
        v.visit(tree)
        v.finalize()
        assert v.is_framework_file is True
        assert 4 in v.framework_decorated_lines


class TestDetectFrameworkUsage:
    def test_decorated_endpoint_confidence_is_zero(self):
        d = Mock()
        d.line = 10
        d.simple_name = "get_users"
        d.type = "function"
        v = FrameworkAwareVisitor()
        v.framework_decorated_lines = {10}
        v.is_framework_file = True
        assert detect_framework_usage(d, visitor=v) == 0

    def test_undecorated_function_in_framework_file_returns_none(self):
        d = Mock()
        d.line = 15
        d.simple_name = "helper_function"
        d.type = "function"
        v = Mock()
        v.framework_decorated_lines = set()
        v.is_framework_file = True
        assert detect_framework_usage(d, visitor=v) is None

    def test_private_function_in_framework_file_returns_none(self):
        d = Mock()
        d.line = 20
        d.simple_name = "_private_function"
        d.type = "function"
        v = Mock()
        v.framework_decorated_lines = set()
        v.is_framework_file = True
        assert detect_framework_usage(d, visitor=v) is None

    def test_non_framework_file_returns_none(self):
        d = Mock()
        d.line = 25
        d.simple_name = "regular_function"
        d.type = "function"
        v = Mock()
        v.framework_decorated_lines = set()
        v.is_framework_file = False
        assert detect_framework_usage(d, visitor=v) is None

    def test_no_visitor_returns_none(self):
        d = Mock()
        assert detect_framework_usage(d, visitor=None) is None

    def test_non_function_in_framework_file_returns_none(self):
        d = Mock()
        d.line = 30
        d.simple_name = "my_variable"
        d.type = "variable"
        v = Mock()
        v.framework_decorated_lines = set()
        v.is_framework_file = True
        assert detect_framework_usage(d, visitor=v) is None


class TestFrameworkPatterns:
    def test_framework_decorators_list(self):
        assert "@*.route" in FRAMEWORK_DECORATORS
        assert "@*.get" in FRAMEWORK_DECORATORS
        assert "@login_required" in FRAMEWORK_DECORATORS

    def test_framework_functions_list(self):
        assert "get" in FRAMEWORK_FUNCTIONS
        assert "post" in FRAMEWORK_FUNCTIONS
        assert "*_queryset" in FRAMEWORK_FUNCTIONS
        assert "get_context_data" in FRAMEWORK_FUNCTIONS

    def test_framework_imports_set(self):
        assert "flask" in FRAMEWORK_IMPORTS
        assert "django" in FRAMEWORK_IMPORTS
        assert "fastapi" in FRAMEWORK_IMPORTS
        assert "pydantic" in FRAMEWORK_IMPORTS


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



import re
from pathlib import Path

PENALTIES = {
    "private_name": 10,
    "dunder_or_magic": 100,
    "in_init_file": 15,
    "dynamic_module": 10,
    "test_related": 100,
    "framework_magic": 40,
}

# timeout const
SUBPROCESS_TIMEOUT = 10
NETWORK_TIMEOUT_SHORT = 5
NETWORK_TIMEOUT_DEFAULT = 15
NETWORK_TIMEOUT_LONG = 30
UPLOAD_TIMEOUT = 60

# safety score const
SAFETY_VERY_HIGH = 0.95
SAFETY_HIGH = 0.90
SAFETY_MEDIUM = 0.85
SAFETY_LOW = 0.70
SAFETY_MINIMAL = 0.60
SAFETY_BUMP = 0.05

# detection thresholds
MIN_SECRET_LENGTH = 16
MIN_LONG_SECRET_LENGTH = 20
ENTROPY_THRESHOLD = 4.5

# magic numbers
MARKREFS_TICK_DEFAULT = 5000
SNIPPET_CONTEXT_LINES = 3

TEST_FILE_RE = re.compile(r"(?:^|[/\\])tests?[/\\]|_test\.py$", re.I)
TEST_IMPORT_RE = re.compile(r"^(pytest|unittest|nose|mock|responses)(\.|$)")
TEST_DECOR_RE = re.compile(
    r"""^(
    pytest\.(fixture|mark) |
    patch(\.|$) |
    responses\.activate |
    freeze_time
)$""",
    re.X,
)

AUTO_CALLED = {
    "__init__",
    "__init__",
    "__new__",
    "__del__",
    "__init_subclass__",
    "__set_name__",
    "__enter__",
    "__exit__",
    "__iter__",
    "__next__",
    "__len__",
    "__getitem__",
    "__setitem__",
    "__delitem__",
    "__contains__",
    "__missing__",
    "__getattr__",
    "__setattr__",
    "__delattr__",
    "__getattribute__",
    "__str__",
    "__repr__",
    "__format__",
    "__bytes__",
    "__hash__",
    "__bool__",
}

TEST_METHOD_PATTERN = re.compile(r"^test_\w+$")

UNITTEST_LIFECYCLE_METHODS = {
    "setUp",
    "tearDown",
    "setUpClass",
    "tearDownClass",
    "setUpModule",
    "tearDownModule",
}

FRAMEWORK_FILE_RE = re.compile(r"(?:views|handlers|endpoints|routes|api)\.py$", re.I)

NON_LIBRARY_DIR_KINDS = {
    "test": "test",
    "tests": "test",
    "__tests__": "test",
    "example": "example",
    "examples": "example",
    "benchmark": "benchmark",
    "benchmarks": "benchmark",
    "bench": "benchmark",
    "docs_src": "example",
    "doc_src": "example",
}

DEFAULT_EXCLUDE_FOLDERS = {
    "__pycache__",
    ".git",
    ".pytest_cache",
    ".mypy_cache",
    ".tox",
    "htmlcov",
    ".coverage",
    "build",
    "dist",
    "*.egg-info",
    "venv",
    ".venv",
    "node_modules",
    ".hg",
    ".svn",
    "vendor",
    ".next",
    ".nuxt",
    ".turbo",
    ".idea",
    ".vscode",
}


def is_test_path(p) -> bool:
    return bool(TEST_FILE_RE.search(str(p)))


def is_framework_path(p) -> bool:
    return bool(FRAMEWORK_FILE_RE.search(str(p)))


def get_non_library_dir_kind(p, project_root=None, extra_dirs=None) -> str | None:
    merged = dict(NON_LIBRARY_DIR_KINDS)
    if extra_dirs:
        for folder, role in extra_dirs.items():
            merged[folder.lower()] = role

    parts = None
    if project_root is not None:
        try:
            rel = Path(p).resolve().relative_to(Path(project_root).resolve())
            parts = rel.parts
        except (TypeError, ValueError, OSError):
            parts = None

    if parts is None:
        parts = str(p).replace("\\", "/").split("/")

    for part in parts:
        kind = merged.get(part.lower())
        if kind:
            return kind

    basename = str(p).replace("\\", "/").rsplit("/", 1)[-1].lower()
    if (
        basename == "conftest.py"
        or basename.startswith("test_")
        or basename.endswith("_test.py")
        or ".test." in basename
        or ".spec." in basename
        or basename.endswith("_test.go")
    ):
        return "test"

    return None


def parse_exclude_folders(
    user_exclude_folders=None, use_defaults=True, include_folders=None
) -> set[str]:
    exclude_folders = set()

    if use_defaults:
        exclude_folders.update(DEFAULT_EXCLUDE_FOLDERS)

    if user_exclude_folders:
        exclude_folders.update(user_exclude_folders)

    if include_folders:
        for folder in include_folders:
            exclude_folders.discard(folder)

    return exclude_folders



from skylos.defend.plugin import DefensePlugin
from skylos.defend.result import DefenseResult
from skylos.discover.integration import LLMIntegration
from skylos.discover.graph import AIIntegrationGraph


class CostControlsPlugin(DefensePlugin):
    id = "cost-controls"
    name = "Cost Controls"
    category = "ops"
    severity = "medium"
    owasp_llm = "LLM10"
    description = (
        "LLM API calls should set max_tokens to control costs "
        "and prevent unbounded token consumption"
    )
    remediation = (
        "Set max_tokens (or max_output_tokens) parameter in LLM API calls "
        "to cap response length and control costs."
    )

    def check(
        self, integration: LLMIntegration, graph: AIIntegrationGraph
    ) -> DefenseResult:
        if integration.has_max_tokens:
            return self._pass(
                integration,
                integration.location,
                "max_tokens set on LLM call",
            )

        return self._fail(
            integration,
            integration.location,
            "No max_tokens set — unbounded token consumption",
        )



import skylos.scale.parallel_static as ps


class DummyFuture:
    def __init__(self, value):
        self._value = value

    def result(self):
        return self._value


class DummyExecutor:
    def __init__(self, max_workers=None):
        self.max_workers = max_workers
        self.futures = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False

    def submit(self, fn, f, mod, extra_visitors, full_scan=True):
        file_str, out = fn(f, mod, extra_visitors, full_scan)
        fut = DummyFuture((file_str, out))
        self.futures.append(fut)
        return fut


def fake_as_completed(futs):
    fs = list(futs)
    fs.reverse()
    for f in fs:
        yield f


def test_parallel_path_preserves_order(monkeypatch, tmp_path):
    monkeypatch.setattr(ps, "ProcessPoolExecutor", DummyExecutor)
    monkeypatch.setattr(ps, "as_completed", fake_as_completed)

    monkeypatch.delenv("PYTEST_CURRENT_TEST", raising=False)

    def fake_proc_file(file_path, mod, extra_visitors=None, full_scan=True):
        return ("ok", str(file_path), mod)

    import skylos.analyzer

    monkeypatch.setattr(skylos.analyzer, "proc_file", fake_proc_file)

    files = [tmp_path / "x.py", tmp_path / "y.py", tmp_path / "z.py"]
    modmap = {files[0]: "mx", files[1]: "my", files[2]: "mz"}

    out = ps.run_proc_file_parallel(files, modmap, jobs=2)

    assert out[0] == ("ok", str(files[0]), "mx")
    assert out[1] == ("ok", str(files[1]), "my")
    assert out[2] == ("ok", str(files[2]), "mz")



import ast
from skylos.visitors.framework_aware import FrameworkAwareVisitor


class TestCLIDecoratorDetection:
    def test_cyclopts_app_command(self):
        code = """
import cyclopts
app = cyclopts.App()

@app.command
def greet(name: str):
    print(f"Hello {name}")
"""
        tree = ast.parse(code)
        visitor = FrameworkAwareVisitor()
        visitor.visit(tree)
        visitor.finalize()

        assert visitor.is_framework_file
        assert 6 in visitor.framework_decorated_lines

    def test_cyclopts_app_default(self):
        code = """
import cyclopts
app = cyclopts.App()

@app.default
def main():
    print("Default command")
"""
        tree = ast.parse(code)
        visitor = FrameworkAwareVisitor()
        visitor.visit(tree)
        visitor.finalize()

        assert visitor.is_framework_file
        assert 6 in visitor.framework_decorated_lines

    def test_typer_app_command(self):
        """@app.command should work for typer too."""
        code = """
import typer
app = typer.Typer()

@app.command()
def hello(name: str):
    typer.echo(f"Hello {name}")
"""
        tree = ast.parse(code)
        visitor = FrameworkAwareVisitor()
        visitor.visit(tree)
        visitor.finalize()

        assert visitor.is_framework_file
        assert 6 in visitor.framework_decorated_lines

    def test_click_command(self):
        """@cli.command should work for click."""
        code = """
import click

@click.group()
def cli():
    pass

@cli.command()
def init():
    click.echo("Initialized")
"""
        tree = ast.parse(code)
        visitor = FrameworkAwareVisitor()
        visitor.visit(tree)
        visitor.finalize()

        assert visitor.is_framework_file
        assert 5 in visitor.framework_decorated_lines
        assert 9 in visitor.framework_decorated_lines

    def test_app_callback(self):
        code = """
import typer
app = typer.Typer()

@app.callback()
def main(verbose: bool = False):
    pass
"""
        tree = ast.parse(code)
        visitor = FrameworkAwareVisitor()
        visitor.visit(tree)
        visitor.finalize()

        assert 6 in visitor.framework_decorated_lines

    def test_generic_command_decorator(self):
        code = """
@cli.command
def foo():
    pass

@main.command()
def bar():
    pass

@something.subcommand
def baz():
    pass
"""
        tree = ast.parse(code)
        visitor = FrameworkAwareVisitor()
        visitor.visit(tree)
        visitor.finalize()

        assert 3 in visitor.framework_decorated_lines
        assert 7 in visitor.framework_decorated_lines
        assert 11 in visitor.framework_decorated_lines



import time
import json
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor, as_completed

from .context import ContextBuilder
from .agents import AgentConfig, create_agent
from .validator import ResultValidator, deduplicate_findings, merge_findings
from .ui import SkylosUI, estimate_cost

from .schemas import Confidence, AnalysisResult
from skylos.llm.graph import CodeGraph


class AnalyzerConfig:
    def __init__(
        self,
        model="gpt-4.1",
        api_key=None,
        temperature=0.1,
        max_tokens=4096,
        enable_security=True,
        enable_quality=True,
        strict_validation=False,
        min_confidence=Confidence.LOW,
        quiet=False,
        json_output=False,
        stream=True,
        parallel=False,
        max_workers=1,
        max_chunk_tokens=1000,
        smart_filter=True,
        complexity_threshold=5,
        batch_functions=True,
        batch_size=10,
    ):
        self.model = model
        self.api_key = api_key
        self.temperature = temperature
        self.max_tokens = max_tokens

        self.enable_security = enable_security
        self.enable_quality = enable_quality

        self.strict_validation = strict_validation
        self.min_confidence = min_confidence

        self.quiet = quiet
        self.json_output = json_output
        self.stream = stream

        self.parallel = parallel
        self.max_workers = max_workers
        self.max_chunk_tokens = max_chunk_tokens

        self.smart_filter = smart_filter
        self.complexity_threshold = complexity_threshold
        self.batch_functions = batch_functions
        self.batch_size = batch_size


class SkylosLLM:
    def __init__(self, config=None):
        self.config = config or AnalyzerConfig()

        self.ui = SkylosUI(quiet=self.config.quiet)
        self.context_builder = ContextBuilder(
            max_context_tokens=self.config.max_chunk_tokens * 2
        )

        self.validator = ResultValidator(
            strict=self.config.strict_validation,
            min_confidence=self.config.min_confidence,
        )

        self.agent_config = AgentConfig()
        self.agent_config.model = self.config.model
        self.agent_config.api_key = self.config.api_key
        self.agent_config.temperature = self.config.temperature
        self.agent_config.max_tokens = self.config.max_tokens
        self.agent_config.stream = self.config.stream

        self._agents = {}

    def _get_agent(self, agent_type):
        if agent_type not in self._agents:
            self._agents[agent_type] = create_agent(agent_type, self.agent_config)
        return self._agents[agent_type]

    def _estimate_complexity(self, node):
        import ast

        complexity = 1
        for child in ast.walk(node):
            if isinstance(
                child,
                (
                    ast.If,
                    ast.While,
                    ast.For,
                    ast.ExceptHandler,
                    ast.With,
                    ast.Assert,
                    ast.comprehension,
                ),
            ):
                complexity += 1
            elif isinstance(child, ast.BoolOp):
                complexity += len(child.values) - 1
        return complexity

    def _should_analyze_function(self, func_name, def_data, graph):
        if not self.config.smart_filter:
            return True

        taint_paths = graph.find_taint_paths(func_name)
        if taint_paths:
            return True

        node = def_data.get("node")
        if node:
            complexity = self._estimate_complexity(node)
            if complexity >= self.config.complexity_threshold:
                return True

        sensitive = [
            "auth",
            "login",
            "password",
            "token",
            "secret",
            "sql",
            "query",
            "execute",
            "eval",
            "exec",
            "shell",
            "command",
            "system",
            "pickle",
            "yaml",
            "upload",
            "file",
            "path",
        ]
        func_lower = func_name.lower()
        for pattern in sensitive:
            if pattern in func_lower:
                return True

        return False

    def _build_batched_context(self, functions_data, graph, file_path, defs_map=None):
        parts = [f"# File: {file_path}\n"]

        for i, (func_name, def_data, taint_paths) in enumerate(functions_data):
            context = graph.get_security_context(func_name, defs_map=defs_map)
            if context:
                parts.append(f"## Function {i + 1}: {func_name}")
                if taint_paths:
                    parts.append(f"# WARNING: {len(taint_paths)} taint flow(s)")
                parts.append(context)
                parts.append("")

        return "\n".join(parts)

    def _analyze_whole_file(
        self,
        source,
        file_path,
        defs_map=None,
        chunk_start_line=1,
        issue_types=None,
        **kwargs,
    ):
        context = self.context_builder.build_analysis_context(
            source, file_path=file_path, defs_map=defs_map
        )

        type_to_agent = {
            "security": "security",
            "quality": "quality",
            "security_audit": "security_audit",
        }

        if not issue_types:
            agent_types = []
            if self.config.enable_security:
                agent_types.append("security")
            if self.config.enable_quality:
                agent_types.append("quality")

            if not agent_types:
                agent_types = ["security_audit"]
        else:
            # Fail fast if caller asks for dead_code through per-file analysis
            for t in issue_types:
                if str(t).lower().strip() == "dead_code":
                    raise ValueError(
                        "Dead code analysis is not a per-file operation. "
                        "Use DeadCodeAgent.verify_candidates() with static "
                        "candidates instead (pipeline Phase 2a)."
                    )

            agent_types = []
            for t in issue_types:
                a = type_to_agent.get(str(t).lower().strip())
                if a:
                    agent_types.append(a)

            if not agent_types:
                agent_types = ["security_audit"]

        all_findings = []

        for agent_type in agent_types:
            agent = self._get_agent(agent_type)

            try:
                if not self.config.quiet:
                    with self.ui.status(f"Analyzing {Path(file_path).name}..."):
                        findings = agent.analyze(
                            source, file_path, defs_map, context=context
                        )
                else:
                    findings = agent.analyze(
                        source, file_path, defs_map, context=context
                    )

            except Exception as e:
                self.ui.print(f"Error analyzing {file_path}: {e}", style="red")
                continue

            if chunk_start_line != 1:
                for f in findings:
                    f.location.line += chunk_start_line - 1

            all_findings.extend(findings)

        return all_findings

    def analyze_file(
        self,
        file_path,
        defs_map=None,
        static_findings=None,
        issue_types=None,
    ):
        file_path = Path(file_path)

        if not file_path.exists():
            return []

        try:
            source = file_path.read_text(encoding="utf-8")
        except Exception as e:
            self.ui.print(f"Error reading {file_path}: {e}", style="red")
            return []

        graph = CodeGraph()
        graph.build(source)

        all_findings = []

        if graph.definitions:
            functions_to_analyze = []
            skipped = 0

            for func_name, def_data in graph.definitions.items():
                if def_data["type"] not in ("function", "method"):
                    continue

                if self._should_analyze_function(func_name, def_data, graph):
                    taint_paths = graph.find_taint_paths(func_name)
                    functions_to_analyze.append((func_name, def_data, taint_paths))
                else:
                    skipped += 1

            if not self.config.quiet and skipped > 0:
                total = len(functions_to_analyze) + skipped
                self.ui.print(
                    f"  Analyzing {len(functions_to_analyze)}/{total} functions (skipped {skipped} low-risk)",
                    style="dim",
                )

            if self.config.batch_functions and len(functions_to_analyze) > 1:
                for i in range(0, len(functions_to_analyze), self.config.batch_size):
                    batch = functions_to_analyze[i : i + self.config.batch_size]
                    batch_context = self._build_batched_context(
                        batch, graph, file_path, defs_map
                    )

                    findings = self._analyze_whole_file(
                        batch_context, str(file_path), defs_map, issue_types=issue_types
                    )

                    for finding in findings:
                        for func_name, def_data, taint_paths in batch:
                            if func_name.split(".")[-1] in str(finding.message):
                                finding.location.line = (
                                    def_data["start"] + finding.location.line
                                )
                                if taint_paths:
                                    for tp in taint_paths:
                                        if tp["sink_type"] in finding.message.lower():
                                            finding.confidence = Confidence.HIGH
                                break

                    all_findings.extend(findings)
            else:
                for func_name, def_data, taint_paths in functions_to_analyze:
                    context = graph.get_security_context(func_name)
                    if not context:
                        continue

                    findings = self._analyze_whole_file(
                        context, str(file_path), defs_map, issue_types=issue_types
                    )

                    if taint_paths:
                        for f in findings:
                            for tp in taint_paths:
                                if tp["sink_type"] in f.message.lower():
                                    f.confidence = Confidence.HIGH

                    start_line = def_data["start"] + 1
                    for f in findings:
                        f.location.line = start_line + f.location.line - 1

                    all_findings.extend(findings)
        else:
            all_findings = self._analyze_whole_file(
                source, str(file_path), defs_map, issue_types=issue_types
            )

        validated, _ = self.validator.validate(all_findings, source, str(file_path))

        if static_findings:
            validated = merge_findings(validated, static_findings, str(file_path))

        validated = deduplicate_findings(validated)

        return validated

    def _count_lines(self, file_path):
        try:
            return len(Path(file_path).read_text(encoding="utf-8").splitlines())
        except Exception:
            return 0

    def _generate_summary(self, result):
        if not result.findings:
            return "No issues found"

        by_severity = {}
        for f in result.findings:
            sev = f.severity.value
            by_severity[sev] = by_severity.get(sev, 0) + 1

        parts = []
        for sev in ["critical", "high", "medium", "low", "info"]:
            if sev in by_severity:
                parts.append(f"{by_severity[sev]} {sev}")

        return f"Found {len(result.findings)} issues: " + ", ".join(parts)

    def analyze_files(
        self,
        files,
        defs_map=None,
        static_findings=None,
        progress_callback=None,
        issue_types=None,
    ):
        start_time = time.time()
        all_findings = []
        total_lines = 0

        files = [Path(f) for f in files]

        if not self.config.quiet:
            self.ui.print_banner()
            tokens, cost = estimate_cost(files, self.config.model)
            self.ui.print(
                f"{len(files)} files, ~{tokens:,} tokens, ~${cost:.4f}", style="dim"
            )

        if self.config.parallel and len(files) > 1:
            with self.ui.create_progress() as progress:
                task = progress.add_task("Analyzing...", total=len(files))

                with ThreadPoolExecutor(
                    max_workers=self.config.max_workers
                ) as executor:
                    future_to_file = {
                        executor.submit(
                            self.analyze_file,
                            f,
                            defs_map,
                            static_findings.get(str(f)) if static_findings else None,
                            issue_types,
                        ): f
                        for f in files
                    }

                    for i, future in enumerate(as_completed(future_to_file)):
                        file = future_to_file[future]
                        try:
                            findings = future.result()
                            all_findings.extend(findings)
                            total_lines += self._count_lines(file)
                        except Exception as e:
                            self.ui.print(f"{file.name}: {e}", style="red")

                        progress.update(
                            task,
                            advance=1,
                            description=f"[{i + 1}/{len(files)}] {file.name}",
                        )
                        if progress_callback:
                            progress_callback(i + 1, len(files), file)
        else:
            with self.ui.create_progress() as progress:
                task = progress.add_task("Analyzing...", total=len(files))

                for i, file in enumerate(files):
                    progress.update(
                        task, description=f"[{i + 1}/{len(files)}] {file.name}"
                    )

                    findings = self.analyze_file(
                        file,
                        defs_map,
                        static_findings.get(str(file)) if static_findings else None,
                        issue_types=issue_types,
                    )
                    all_findings.extend(findings)
                    total_lines += self._count_lines(file)

                    progress.update(task, advance=1)
                    if progress_callback:
                        progress_callback(i + 1, len(files), file)

        elapsed_ms = int((time.time() - start_time) * 1000)

        result = AnalysisResult(
            findings=all_findings,
            files_analyzed=len(files),
            total_lines=total_lines,
            analysis_time_ms=elapsed_ms,
            model_used=self.config.model,
        )
        result.summary = self._generate_summary(result)

        return result

    def analyze_project(
        self,
        project_path,
        exclude_folders=None,
        defs_map=None,
        static_findings=None,
        issue_types=None,
    ):
        project_path = Path(project_path)

        if not project_path.exists():
            return AnalysisResult(summary="Project path not found")

        exclude = set(exclude_folders or [])
        exclude.update(
            {
                "__pycache__",
                ".git",
                ".venv",
                "venv",
                "node_modules",
                ".pytest_cache",
                ".mypy_cache",
                "build",
                "dist",
                ".tox",
                ".eggs",
            }
        )

        files = []
        for f in project_path.rglob("*.py"):
            skip = False
            for part in f.parts:
                if part in exclude or part.endswith(".egg-info"):
                    skip = True
                    break
            if not skip:
                files.append(f)

        if not files:
            return AnalysisResult(summary="No Python files found")

        return self.analyze_files(
            files, defs_map, static_findings, issue_types=issue_types
        )

    def fix_issue(self, file_path, issue_line, issue_message, defs_map=None):
        file_path = Path(file_path)
        if not file_path.exists():
            return None

        try:
            source = file_path.read_text(encoding="utf-8")
        except Exception:
            return None

        context = self.context_builder.build_fix_context(
            source, str(file_path), issue_line, issue_message, defs_map
        )

        fixer = self._get_agent("fixer")
        return fixer.fix(
            source, str(file_path), issue_line, issue_message, defs_map, context
        )

    def print_results(self, result, format="table", output_file=None):
        if format == "json":
            output = json.dumps(result.to_dict(), indent=2)
            if output_file:
                Path(output_file).write_text(output)
            elif not self.config.quiet:
                print(output)
            return

        if format == "sarif":
            output = json.dumps(result.to_sarif(), indent=2)
            if output_file:
                Path(output_file).write_text(output)
            elif not self.config.quiet:
                print(output)
            return

        if self.config.quiet:
            return

        if format == "tree":
            self.ui.print_findings_tree(result.findings)
        else:
            self.ui.print_findings_table(result.findings)

        self.ui.print_summary(result)


def analyze(path, model="gpt-4.1", issue_types=None, **kwargs):
    config = AnalyzerConfig(model=model, **kwargs)
    analyzer = SkylosLLM(config)

    path = Path(path)
    if path.is_file():
        findings = analyzer.analyze_file(path, issue_types=issue_types)
        return AnalysisResult(findings=findings, files_analyzed=1)
    else:
        return analyzer.analyze_project(path, issue_types=issue_types)


def audit(path, model="gpt-4.1", **kwargs):
    return analyze(path, model=model, issue_types=["security_audit"], **kwargs)


def fix(file_path, line, message, model="gpt-4.1"):
    config = AnalyzerConfig(model=model)
    analyzer = SkylosLLM(config)
    return analyzer.fix_issue(file_path, line, message)



import ast

XSS_SANITIZERS = {
    "html.escape",
    "markupsafe.escape",
    "markupsafe.Markup.escape",
    "Markup.escape",
    "bleach.clean",
    "bleach.linkify",
    "cgi.escape",
}

CMD_SANITIZERS = {
    "shlex.quote",
    "shlex.join",
}

URL_SANITIZERS = {
    "urllib.parse.quote",
    "urllib.parse.quote_plus",
}

PATH_SANITIZERS = {
    "os.path.basename",
    "pathlib.PurePath.name",
}


def _resolve_call_name(node: ast.Call):
    func = node.func
    parts = []
    while isinstance(func, ast.Attribute):
        parts.append(func.attr)
        func = func.value
    if isinstance(func, ast.Name):
        parts.append(func.id)
        parts.reverse()
        return ".".join(parts)
    return None


class TaintVisitor(ast.NodeVisitor):
    def __init__(self, file_path, findings, sanitizers=None):
        self.file_path = file_path
        self.findings = findings
        self.env_stack = [{}]
        self.sources = {"input", "request"}
        self.request_obj = "request"
        self._symbol_stack = ["<module>"]
        self.sanitizers = sanitizers or set()

    def _current_symbol(self):
        if self._symbol_stack:
            return self._symbol_stack[-1]
        else:
            return "<module>"

    def _push(self):
        self.env_stack.append({})

    def _pop(self):
        if self.env_stack:
            self.env_stack.pop()

    def _set(self, name, tainted):
        if not self.env_stack:
            self._push()
        self.env_stack[-1][name] = bool(tainted)

    def _get(self, name):
        for env in reversed(self.env_stack):
            if name in env:
                return env[name]
        return False

    def _taint_params(self, fn: ast.AST):
        args = []
        if hasattr(fn, "args") and fn.args:
            args.extend(getattr(fn.args, "posonlyargs", []) or [])
            args.extend(getattr(fn.args, "args", []) or [])
            args.extend(getattr(fn.args, "kwonlyargs", []) or [])

            if fn.args.vararg:
                args.append(fn.args.vararg)
            if fn.args.kwarg:
                args.append(fn.args.kwarg)

        for a in args:
            name = getattr(a, "arg", None)
            if not name:
                continue
            if name in ("self", "cls"):
                continue
            self._set(name, True)

    def is_tainted(self, node):
        if node is None:
            return False

        if isinstance(node, ast.JoinedStr):
            return any(
                isinstance(v, ast.FormattedValue) and self.is_tainted(v.value)
                for v in node.values
            )
        if isinstance(node, ast.BinOp) and isinstance(node.op, (ast.Add, ast.Mod)):
            return self.is_tainted(node.left) or self.is_tainted(node.right)
        if (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Attribute)
            and node.func.attr == "format"
        ):
            return True

        if (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Name)
            and node.func.id in self.sources
        ):
            return True

        if isinstance(node, (ast.Attribute, ast.Subscript)):
            base = node.value
            while isinstance(base, (ast.Attribute, ast.Subscript)):
                base = base.value
            if isinstance(base, ast.Name) and base.id == self.request_obj:
                return True
            return self.is_tainted(node.value)

        if isinstance(node, ast.Name):
            return self._get(node.id)

        if isinstance(node, ast.Call):
            if self.sanitizers:
                call_name = _resolve_call_name(node)
                if call_name and call_name in self.sanitizers:
                    return False

            if isinstance(node.func, ast.Attribute):
                if self.is_tainted(node.func.value):
                    return True

            if any(self.is_tainted(arg) for arg in node.args):
                return True

            if any(self.is_tainted(k.value) for k in node.keywords):
                return True

        return False

    def generic_visit(self, node):
        for field, value in ast.iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, ast.AST):
                        self.visit(item)
            elif isinstance(value, ast.AST):
                self.visit(value)

    def visit_FunctionDef(self, node):
        self._push()
        self._symbol_stack.append(node.name)
        self._taint_params(node)
        self.generic_visit(node)
        self._symbol_stack.pop()
        self._pop()

    def visit_AsyncFunctionDef(self, node):
        self._push()
        self._symbol_stack.append(node.name)
        self._taint_params(node)
        self.generic_visit(node)
        self._symbol_stack.pop()
        self._pop()

    def visit_ClassDef(self, node):
        self._symbol_stack.append(node.name)
        self.generic_visit(node)
        self._symbol_stack.pop()

    def visit_Assign(self, node):
        t = self.is_tainted(node.value)
        for tgt in node.targets:
            if isinstance(tgt, ast.Name):
                self._set(tgt.id, t)
        self.generic_visit(node)

    def visit_AnnAssign(self, node):
        if node.value:
            t = self.is_tainted(node.value)
            if isinstance(node.target, ast.Name):
                self._set(node.target.id, t)
        self.generic_visit(node)



from __future__ import annotations

import json
import hashlib
import logging
import os
import re
from datetime import datetime, timezone
from pathlib import Path
from typing import Any

from skylos_mcp.auth import (
    initialize_auth,
    check_tool_access,
    deduct_credits,
)


def _lazy_analyze():
    from skylos.analyzer import analyze as run_analyze

    return run_analyze


def _lazy_constants():
    from skylos.constants import parse_exclude_folders, DEFAULT_EXCLUDE_FOLDERS

    return parse_exclude_folders, DEFAULT_EXCLUDE_FOLDERS


logger = logging.getLogger("skylos-mcp")


RESULTS_DIR = Path(
    os.getenv("SKYLOS_MCP_RESULTS_DIR", Path.home() / ".skylos" / "mcp_results")
)
RESULTS_DIR.mkdir(parents=True, exist_ok=True)

_results_cache: dict[str, dict[str, Any]] = {}


def _store_result(result: dict, tool: str, path: str) -> str:
    ts = datetime.now(timezone.utc).isoformat()
    run_id = hashlib.sha256(f"{ts}-{tool}-{path}".encode()).hexdigest()[:12]

    envelope = {
        "run_id": run_id,
        "tool": tool,
        "path": path,
        "timestamp": ts,
        "result": result,
    }

    _results_cache[run_id] = envelope
    _results_cache["latest"] = envelope

    try:
        (RESULTS_DIR / f"{run_id}.json").write_text(json.dumps(envelope, indent=2))
        (RESULTS_DIR / "latest.json").write_text(json.dumps(envelope, indent=2))
    except OSError as exc:
        logger.warning("Could not persist result to disk: %s", exc)

    return run_id


def _load_result(run_id: str) -> dict | None:
    if run_id in _results_cache:
        return _results_cache[run_id]

    disk = RESULTS_DIR / f"{run_id}.json"
    if disk.exists():
        data = json.loads(disk.read_text())
        _results_cache[run_id] = data
        return data
    return None


def _list_runs() -> list[dict]:
    seen = set()
    runs = []

    for f in sorted(
        RESULTS_DIR.glob("*.json"), key=lambda p: p.stat().st_mtime, reverse=True
    ):
        if f.stem == "latest":
            continue
        try:
            data = json.loads(f.read_text())
            rid = data["run_id"]
            if rid not in seen:
                seen.add(rid)
                runs.append(
                    {
                        "run_id": rid,
                        "tool": data.get("tool"),
                        "path": data.get("path"),
                        "timestamp": data.get("timestamp"),
                    }
                )
        except (OSError, json.JSONDecodeError, KeyError):
            continue

    for rid, data in _results_cache.items():
        if rid == "latest" or rid in seen:
            continue
        seen.add(rid)
        runs.append(
            {
                "run_id": rid,
                "tool": data.get("tool"),
                "path": data.get("path"),
                "timestamp": data.get("timestamp"),
            }
        )

    return runs


def _validate_code_change_impl(
    diff: str,
    path: str = ".",
    policy: str | None = None,
) -> dict:
    """Core logic for validate_code_change, extracted for testability."""
    from skylos.rules.quality.regression import detect_security_regressions
    from skylos.rules.secrets import (
        PROVIDER_PATTERNS,
        GENERIC_VALUE,
        SAFE_TEST_HINTS,
        _entropy,
        DEFAULT_MIN_ENTROPY,
    )

    all_findings: list[dict] = []

    file_chunks: list[tuple[str, str]] = []
    current_file = None
    current_lines: list[str] = []

    for line in diff.splitlines():
        if line.startswith("+++ b/"):
            if current_file and current_lines:
                file_chunks.append((current_file, "\n".join(current_lines)))
            current_file = line[6:]
            current_lines = [line]
        elif line.startswith("--- a/") or line.startswith("--- /dev/null"):
            current_lines.append(line)
        elif current_file is not None:
            current_lines.append(line)

    if current_file and current_lines:
        file_chunks.append((current_file, "\n".join(current_lines)))

    for file_path_chunk, chunk_text in file_chunks:
        regressions = detect_security_regressions(chunk_text, file_path_chunk)
        all_findings.extend(regressions)

    dangerous_calls = {
        "eval(": ("SKY-D201", "HIGH", "Use of eval()"),
        "exec(": ("SKY-D202", "HIGH", "Use of exec()"),
        "os.system(": ("SKY-D203", "CRITICAL", "Use of os.system()"),
        "pickle.load(": ("SKY-D204", "CRITICAL", "Untrusted deserialization via pickle.load"),
        "pickle.loads(": ("SKY-D205", "CRITICAL", "Untrusted deserialization via pickle.loads"),
        "yaml.load(": ("SKY-D206", "HIGH", "yaml.load without SafeLoader"),
        "marshal.loads(": ("SKY-D233", "CRITICAL", "Untrusted deserialization via marshal.loads"),
        "__import__(": ("SKY-D240", "HIGH", "Dynamic import via __import__()"),
        "compile(": ("SKY-D241", "MEDIUM", "Dynamic code compilation"),
    }

    sql_injection_re = re.compile(
        r"""(?:execute|cursor\.execute|raw|text)\(\s*(?:f['"']|['"].*%[sd]|['"].*\+|.*\.format\()""",
    )

    for file_path_chunk, chunk_text in file_chunks:
        line_no = 0
        for raw_line in chunk_text.splitlines():
            if raw_line.startswith("@@"):
                m = re.match(r"@@ -\d+(?:,\d+)? \+(\d+)", raw_line)
                if m:
                    line_no = int(m.group(1)) - 1
                continue
            if raw_line.startswith("+") and not raw_line.startswith("+++"):
                line_no += 1
                added = raw_line[1:]
                stripped = added.strip()

                for pattern, (rule_id, severity, msg) in dangerous_calls.items():
                    if pattern in stripped:
                        all_findings.append({
                            "rule_id": rule_id,
                            "kind": "dangerous_pattern",
                            "severity": severity,
                            "message": f"Dangerous pattern in added code: {msg}",
                            "file": file_path_chunk,
                            "line": line_no,
                            "col": 0,
                        })

                if sql_injection_re.search(stripped):
                    all_findings.append({
                        "rule_id": "SKY-D220",
                        "kind": "dangerous_pattern",
                        "severity": "CRITICAL",
                        "message": "Potential SQL injection in added code: string interpolation in query",
                        "file": file_path_chunk,
                        "line": line_no,
                        "col": 0,
                    })

                for provider, regex in PROVIDER_PATTERNS:
                    if regex.search(stripped):
                        all_findings.append({
                            "rule_id": "SKY-S101",
                            "kind": "secret",
                            "severity": "CRITICAL",
                            "message": f"Secret detected in added code: {provider} token/key",
                            "file": file_path_chunk,
                            "line": line_no,
                            "col": 0,
                        })

                for m_generic in GENERIC_VALUE.finditer(stripped):
                    val = m_generic.group("val") or m_generic.group("bare")
                    if val and _entropy(val) >= DEFAULT_MIN_ENTROPY:
                        if not any(hint in val.lower() for hint in SAFE_TEST_HINTS):
                            all_findings.append({
                                "rule_id": "SKY-S101",
                                "kind": "secret",
                                "severity": "HIGH",
                                "message": "Possible secret/credential in added code",
                                "file": file_path_chunk,
                                "line": line_no,
                                "col": 0,
                            })
            elif not raw_line.startswith("-"):
                line_no += 1

    if policy:
        policy_path = Path(policy) if not Path(policy).is_absolute() else Path(policy)
        if not policy_path.exists():
            target_dir = Path(path).resolve()
            policy_path = target_dir / policy

    status = "fail" if all_findings else "pass"
    counts: dict[str, int] = {}
    for f in all_findings:
        kind = f.get("kind", "unknown")
        counts[kind] = counts.get(kind, 0) + 1

    parts = []
    for kind, count in sorted(counts.items()):
        label = kind.replace("_", " ")
        parts.append(f"{count} {label}{'s' if count != 1 else ''}")
    summary_text = ", ".join(parts) + " found" if parts else "No issues found"

    return {
        "status": status,
        "findings": all_findings,
        "summary": summary_text,
    }


def _get_security_context_impl(path: str) -> dict:
    """Core logic for get_security_context, extracted for testability."""
    target = Path(path).resolve()
    if not target.exists():
        return {"error": f"Path does not exist: {path}"}

    context: dict[str, Any] = {
        "project_path": str(target),
        "frameworks": [],
        "auth_patterns": [],
        "security_headers": [],
        "rate_limiting": [],
        "input_validation": [],
        "policy": None,
    }

    framework_indicators = {
        "Django": [
            "manage.py", "settings.py", "django.conf", "urls.py",
        ],
        "Flask": [
            "app.py", "wsgi.py",
        ],
        "FastAPI": [
            "main.py", "app.py",
        ],
        "Express": [
            "app.js", "server.js", "index.js",
        ],
        "Next.js": [
            "next.config.js", "next.config.mjs", "next.config.ts",
        ],
    }

    framework_imports = {
        "Django": re.compile(r"(?:from django|import django)"),
        "Flask": re.compile(r"(?:from flask |import flask)"),
        "FastAPI": re.compile(r"(?:from fastapi |import fastapi|FastAPI\(\))"),
        "Express": re.compile(r"""(?:require\(['"]express['"]\)|from ['"]express['"])"""),
        "Next.js": re.compile(r"""(?:from ['"]next/|@next/)"""),
    }

    scan_files: list[Path] = []
    scan_extensions = {".py", ".js", ".ts", ".jsx", ".tsx", ".mjs"}
    scan_names = set()
    for indicators in framework_indicators.values():
        scan_names.update(indicators)

    for item in target.rglob("*"):
        parts_item = item.parts
        if any(skip in parts_item for skip in ("node_modules", ".git", "__pycache__", "venv", ".venv", "env")):
            continue
        if item.is_file():
            if item.name in scan_names or item.suffix in scan_extensions:
                scan_files.append(item)
        if len(scan_files) > 500:
            break

    detected_frameworks: set[str] = set()
    auth_patterns: set[str] = set()
    security_headers: set[str] = set()
    rate_limiting: set[str] = set()
    validation: set[str] = set()

    auth_decorator_re = re.compile(
        r"@(?:login_required|require_auth|requires_auth|authenticated|"
        r"permission_required|jwt_required|token_required|permissions_required)"
    )
    auth_depends_re = re.compile(
        r"Depends\((?:get_current_user|get_current_active_user|require_admin|verify_token)\)"
    )
    auth_middleware_re = re.compile(
        r"(?:AuthenticationMiddleware|SessionMiddleware|JWTAuthentication|"
        r"OAuth2PasswordBearer|HTTPBearer)"
    )

    header_names = {
        "X-Content-Type-Options", "X-Frame-Options",
        "Content-Security-Policy", "Strict-Transport-Security",
        "X-XSS-Protection", "Referrer-Policy", "Permissions-Policy",
    }
    header_middleware_re = re.compile(r"(?:SecurityMiddleware|helmet\(|secure_headers)")

    rate_limit_re = re.compile(
        r"(?:@(?:rate_limit|ratelimit|throttle|limiter\.limit)|"
        r"slowapi|RateLimitMiddleware|Throttle)"
    )

    validation_re = re.compile(
        r"(?:@(?:validate|validator|field_validator|validates)|"
        r"BaseModel|Pydantic|Schema\(|marshmallow|cerberus|"
        r"wtforms|FlaskForm|Serializer)"
    )

    for fpath in scan_files:
        try:
            content = fpath.read_text(errors="ignore")
        except OSError:
            continue

        for fw, regex in framework_imports.items():
            if regex.search(content):
                detected_frameworks.add(fw)

        for m in auth_decorator_re.finditer(content):
            auth_patterns.add(m.group(0))
        for m in auth_depends_re.finditer(content):
            auth_patterns.add(m.group(0))
        for m in auth_middleware_re.finditer(content):
            auth_patterns.add(m.group(0))

        for header in header_names:
            if header in content:
                security_headers.add(header)
        if header_middleware_re.search(content):
            security_headers.add("security_middleware_detected")

        for m in rate_limit_re.finditer(content):
            rate_limiting.add(m.group(0))

        for m in validation_re.finditer(content):
            validation.add(m.group(0))

    context["frameworks"] = sorted(detected_frameworks)
    context["auth_patterns"] = sorted(auth_patterns)
    context["security_headers"] = sorted(security_headers)
    context["rate_limiting"] = sorted(rate_limiting)
    context["input_validation"] = sorted(validation)

    for policy_name in (".skylos.yml", ".skylos.yaml", "skylos.yml", "skylos.yaml"):
        policy_path = target / policy_name
        if policy_path.exists():
            try:
                import yaml
                context["policy"] = yaml.safe_load(policy_path.read_text())
            except Exception:
                context["policy"] = f"Found {policy_name} but could not parse"
            break

    return context


def _run_analysis(
    path: str,
    confidence: int = 60,
    exclude_folders: list[str] | None = None,
    enable_secrets: bool = False,
    enable_danger: bool = False,
    enable_quality: bool = False,
) -> dict:
    analyze = _lazy_analyze()
    parse_exclude_folders, _ = _lazy_constants()

    if exclude_folders is None:
        exclude_folders = list(parse_exclude_folders())

    result_json = analyze(
        path,
        conf=confidence,
        exclude_folders=exclude_folders,
        enable_secrets=enable_secrets,
        enable_danger=enable_danger,
        enable_quality=enable_quality,
    )
    return json.loads(result_json)


def _make_summary(result: dict, focus: str | None = None) -> dict:
    summary = result.get("analysis_summary", {})
    out: dict[str, Any] = {"analysis_summary": summary}

    if focus is None or focus == "dead_code":
        for key in [
            "unused_functions",
            "unused_imports",
            "unused_classes",
            "unused_variables",
            "unused_parameters",
            "unused_files",
        ]:
            items = result.get(key, [])
            if items:
                out[key] = items

    if focus is None or focus == "security":
        if result.get("danger"):
            out["danger"] = result["danger"]

    if focus is None or focus == "quality":
        if result.get("quality"):
            out["quality"] = result["quality"]
        if result.get("circular_dependencies"):
            out["circular_dependencies"] = result["circular_dependencies"]
        if result.get("custom_rules"):
            out["custom_rules"] = result["custom_rules"]

    if focus is None or focus == "secrets":
        if result.get("secrets"):
            out["secrets"] = result["secrets"]

    return out


def _gate(tool_name: str) -> str | None:
    allowed, err = check_tool_access(tool_name)
    if not allowed:
        return json.dumps({"error": err})

    ok, credit_err = deduct_credits(tool_name)
    if not ok:
        return json.dumps({"error": credit_err})

    return None


def _register_tools(mcp):
    """Register all MCP tools and resources. Called inside main() after FastMCP is created."""

    ## all of these look like dead but they're all registered inside `_register_tools()` which is
    ## called from main() .. please ignore the "unused function" warnings for these --- IGNORE ---
    @mcp.tool()
    def analyze(
        path: str,
        confidence: int = 60,
        exclude_folders: list[str] | None = None,
    ) -> str:
        gate_err = _gate("analyze")
        if gate_err:
            return gate_err

        result = _run_analysis(
            path,
            confidence=confidence,
            exclude_folders=exclude_folders,
            enable_secrets=False,
            enable_danger=False,
            enable_quality=False,
        )
        summary = _make_summary(result, focus="dead_code")
        run_id = _store_result(result, "analyze", path)
        summary["_run_id"] = run_id
        return json.dumps(summary, indent=2)

    @mcp.tool()
    def security_scan(
        path: str,
        confidence: int = 60,
        exclude_folders: list[str] | None = None,
    ) -> str:
        gate_err = _gate("security_scan")
        if gate_err:
            return gate_err

        result = _run_analysis(
            path,
            confidence=confidence,
            exclude_folders=exclude_folders,
            enable_danger=True,
        )
        summary = _make_summary(result, focus="security")
        run_id = _store_result(result, "security_scan", path)
        summary["_run_id"] = run_id
        return json.dumps(summary, indent=2)

    @mcp.tool()
    def quality_check(
        path: str,
        confidence: int = 60,
        exclude_folders: list[str] | None = None,
    ) -> str:
        gate_err = _gate("quality_check")
        if gate_err:
            return gate_err

        result = _run_analysis(
            path,
            confidence=confidence,
            exclude_folders=exclude_folders,
            enable_quality=True,
        )
        summary = _make_summary(result, focus="quality")
        run_id = _store_result(result, "quality_check", path)
        summary["_run_id"] = run_id
        return json.dumps(summary, indent=2)

    @mcp.tool()
    def secrets_scan(
        path: str,
        confidence: int = 60,
        exclude_folders: list[str] | None = None,
    ) -> str:
        gate_err = _gate("secrets_scan")
        if gate_err:
            return gate_err

        result = _run_analysis(
            path,
            confidence=confidence,
            exclude_folders=exclude_folders,
            enable_secrets=True,
        )
        summary = _make_summary(result, focus="secrets")
        run_id = _store_result(result, "secrets_scan", path)
        summary["_run_id"] = run_id
        return json.dumps(summary, indent=2)

    @mcp.tool()
    def remediate(
        path: str,
        max_fixes: int = 5,
        dry_run: bool = True,
        model: str = "gpt-4.1",
        test_cmd: str | None = None,
        severity: str | None = None,
    ) -> str:
        gate_err = _gate("remediate")
        if gate_err:
            return gate_err

        try:
            from skylos.llm.orchestrator import RemediationAgent

            agent = RemediationAgent(
                model=model,
                test_cmd=test_cmd,
                severity_filter=severity,
            )
            summary = agent.run(
                path,
                dry_run=dry_run,
                max_fixes=max_fixes,
                quiet=True,
            )
            return json.dumps(summary, indent=2)
        except Exception as e:
            return json.dumps({"error": str(e)})

    @mcp.tool()
    def verify_dead_code(
        path: str,
        confidence: int = 60,
        model: str = "gpt-4.1",
        max_verify: int = 30,
        max_challenge: int = 10,
        exclude_folders: str | None = None,
    ) -> str:
        gate_err = _gate("verify_dead_code")
        if gate_err:
            return gate_err

        try:
            run_analyze = _lazy_analyze()
            parse_exclude_folders, _ = _lazy_constants()

            excl = list(parse_exclude_folders(use_defaults=True))
            if exclude_folders:
                for f in exclude_folders.split(","):
                    f = f.strip()
                    if f:
                        excl.append(f)

            raw = run_analyze(str(path), conf=confidence, exclude_folders=excl)
            static_result = json.loads(raw) if isinstance(raw, str) else raw

            all_findings = []
            for key in (
                "unused_functions",
                "unused_classes",
                "unused_variables",
                "unused_imports",
            ):
                all_findings.extend(static_result.get(key, []))

            defs_map = static_result.get("definitions", {})

            if not all_findings:
                return json.dumps(
                    {"message": "No dead code findings to verify", "stats": {}}
                )

            from skylos.llm.verify_orchestrator import run_verification

            result = run_verification(
                findings=all_findings,
                defs_map=defs_map,
                project_root=str(path),
                model=model,
                max_verify=max_verify,
                max_challenge=max_challenge,
                quiet=True,
            )

            _store_result(result, "verify_dead_code", path)
            return json.dumps(result, indent=2, default=str)

        except Exception as e:
            return json.dumps({"error": str(e)})

    @mcp.tool()
    def provenance_scan(path: str, diff_base: str | None = None) -> str:
        gate_err = _gate("provenance_scan")
        if gate_err:
            return gate_err

        try:
            from skylos.provenance import analyze_provenance
            from skylos.api import get_git_root

            target = os.path.abspath(path)
            git_root = get_git_root() or target
            report = analyze_provenance(git_root, base_ref=diff_base)
            result = report.to_dict()
            _store_result(result, "provenance_scan", path)
            return json.dumps(result, indent=2)
        except Exception as e:
            return json.dumps({"error": str(e)})

    @mcp.tool()
    def generate_fix(
        path: str,
        mode: str = "delete",
        min_safety: float = 0.0,
        apply: bool = False,
    ) -> str:
        gate_err = _gate("generate_fix")
        if gate_err:
            return gate_err

        try:
            from skylos.analyzer import analyze as run_static
            from skylos.dead_code import collect_dead_code_findings
            from skylos.fixgen import (
                generate_removal_plan,
                generate_unified_diff,
                apply_patches,
                validate_patches,
                generate_fix_summary,
            )
            from skylos.grep_verify import grep_verify_findings

            target = os.path.abspath(path)
            raw = run_static(
                target,
                conf=60,
                enable_danger=False,
                enable_quality=False,
                enable_secrets=False,
            )
            static_result = json.loads(raw) if isinstance(raw, str) else raw
            all_findings = collect_dead_code_findings(static_result)
            defs_map = static_result.get("definitions", {})

            # use grep verification to filter
            verdicts = grep_verify_findings(all_findings, target, time_budget=30.0)
            dead_findings = [
                f
                for f in all_findings
                if f.get("full_name", f.get("name", "")) not in verdicts
            ]

            patches = generate_removal_plan(
                dead_findings,
                defs_map,
                target,
                mode=mode,
                min_safety=min_safety,
            )
            errors = validate_patches(patches, target)
            diff = generate_unified_diff(patches, target)
            summary = generate_fix_summary(patches)

            result = {
                "patches": len(patches),
                "summary": summary,
                "errors": errors,
                "diff": diff,
                "applied": False,
            }

            if apply and not errors:
                apply_patches(patches, target, dry_run=False)
                result["applied"] = True

            _store_result(result, "generate_fix", path)
            return json.dumps(result, indent=2)
        except Exception as e:
            return json.dumps({"error": str(e)})

    @mcp.tool()
    def learn_triage(
        path: str,
        action_id: str,
        action: str,
    ) -> str:
        gate_err = _gate("learn_triage")
        if gate_err:
            return gate_err

        try:
            from skylos.agent_service import AgentServiceController

            controller = AgentServiceController(os.path.abspath(path))
            result = controller.learn_triage(action_id, action)
            _store_result(result, "learn_triage", path)
            return json.dumps(result, indent=2)
        except Exception as e:
            return json.dumps({"error": str(e)})

    @mcp.tool()
    def get_triage_suggestions(
        path: str,
    ) -> str:

        gate_err = _gate("get_triage_suggestions")
        if gate_err:
            return gate_err

        try:
            from skylos.agent_service import AgentServiceController

            controller = AgentServiceController(os.path.abspath(path))
            result = controller.get_suggestions()
            _store_result(result, "get_triage_suggestions", path)
            return json.dumps(result, indent=2)
        except Exception as e:
            return json.dumps({"error": str(e)})

    @mcp.tool()
    def validate_code_change(
        diff: str,
        path: str = ".",
        policy: str | None = None,
    ) -> str:
        """Validate a code diff for security regressions and issues before it lands.

        Takes a unified diff and checks for:
        - Security control regressions (auth, CSRF, TLS, rate limiting removal)
        - New dangerous patterns (eval, exec, SQL injection, etc.)
        - Secrets in added code
        - AI defense issues in added code

        Returns pass/fail with findings.
        """
        gate_err = _gate("validate_code_change")
        if gate_err:
            return gate_err

        try:
            result = _validate_code_change_impl(diff, path, policy)
            _store_result(result, "validate_code_change", path)
            return json.dumps(result, indent=2)
        except Exception as e:
            return json.dumps({"error": str(e)})

    @mcp.tool()
    def get_security_context(
        path: str,
    ) -> str:
        gate_err = _gate("get_security_context")
        if gate_err:
            return gate_err

        try:
            result = _get_security_context_impl(path)
            if "error" in result:
                return json.dumps(result)
            _store_result(result, "get_security_context", path)
            return json.dumps(result, indent=2)
        except Exception as e:
            return json.dumps({"error": str(e)})

    @mcp.tool()
    def city_topology(
        path: str,
        confidence: int = 60,
        exclude_folders: list[str] | None = None,
        filter_district: str | None = None,
        filter_dead_only: bool = False,
        min_complexity: int = 0,
    ) -> str:
        """Generate Code City topology from codebase analysis.

        Returns a 3D city representation where directories are districts,
        files are blocks, and functions/classes are buildings.
        Building height = LOC, color = complexity, dead = abandoned.

        Filter params let agents query slices of the city.
        """
        gate_err = _gate("city_topology")
        if gate_err:
            return gate_err

        try:
            from skylos.city import generate_topology

            result = _run_analysis(
                path,
                confidence=confidence,
                exclude_folders=exclude_folders,
                enable_quality=True,
            )
            topology = generate_topology(result)

            # Apply filters
            if filter_district:
                topology["districts"] = [
                    d for d in topology["districts"]
                    if filter_district in d["name"]
                ]

            if filter_dead_only:
                for district in topology["districts"]:
                    for block in district["blocks"]:
                        block["buildings"] = [
                            b for b in block["buildings"] if b["dead"]
                        ]
                    district["blocks"] = [
                        b for b in district["blocks"] if b["buildings"]
                    ]
                topology["districts"] = [
                    d for d in topology["districts"] if d["blocks"]
                ]

            if min_complexity > 0:
                for district in topology["districts"]:
                    for block in district["blocks"]:
                        block["buildings"] = [
                            b for b in block["buildings"]
                            if b["complexity"] >= min_complexity
                        ]
                    district["blocks"] = [
                        b for b in district["blocks"] if b["buildings"]
                    ]
                topology["districts"] = [
                    d for d in topology["districts"] if d["blocks"]
                ]

            run_id = _store_result(topology, "city_topology", path)
            topology["_run_id"] = run_id
            return json.dumps(topology, indent=2)
        except Exception as e:
            return json.dumps({"error": str(e)})

    @mcp.resource("skylos://results/latest")
    def get_latest_result() -> str:
        data = _load_result("latest")
        if data is None:
            return json.dumps({"error": "No analysis has been run yet."})
        return json.dumps(data, indent=2)

    @mcp.resource("skylos://results/{run_id}")
    def get_result_by_id(run_id: str) -> str:
        data = _load_result(run_id)
        if data is None:
            return json.dumps({"error": f"Run '{run_id}' not found."})
        return json.dumps(data, indent=2)

    @mcp.resource("skylos://results")
    def list_results() -> str:
        return json.dumps(_list_runs(), indent=2)


def main():
    try:
        from mcp.server.fastmcp import FastMCP

        initialize_auth()

        transport = os.getenv("MCP_TRANSPORT", "stdio")

        if transport in ("sse", "streamable-http"):
            host = os.getenv("MCP_BIND", "127.0.0.1")
            port = int(os.getenv("PORT", "8080"))
            mcp_server = FastMCP(name="skylos", host=host, port=port)
        else:
            mcp_server = FastMCP(name="skylos")

        _register_tools(mcp_server)
        mcp_server.run(transport=transport)
    except Exception:
        import sys
        import traceback

        traceback.print_exc(file=sys.stderr)
        raise


if __name__ == "__main__":
    main()



from __future__ import annotations

import json
import os
import re
import subprocess

from rich.console import Console

from skylos.rules.quality.regression import detect_security_regressions

console = Console()


def run_pr_review(
    results: dict,
    *,
    pr_number: int | None = None,
    repo: str | None = None,
    summary_only: bool = False,
    max_comments: int = 25,
    diff_base: str = "origin/main",
    grade: dict | None = None,
    previous_grade: dict | None = None,
    llm_findings: list[dict] | None = None,
) -> None:
    pr_number = pr_number or _detect_pr_number()
    repo = repo or os.environ.get("GITHUB_REPOSITORY")

    if not pr_number:
        console.print(
            "[yellow]Could not detect PR number. Use --pr to specify.[/yellow]"
        )
        return

    if not repo:
        console.print("[yellow]Could not detect repo. Use --repo to specify.[/yellow]")
        return

    if not _gh_available():
        console.print(
            "[bold red]gh CLI not found. Install: https://cli.github.com[/bold red]"
        )
        return

    if grade and previous_grade is None:
        previous_grade = _fetch_previous_grade(repo, diff_base)

    all_findings = _flatten_findings(results)

    if llm_findings:
        all_findings = _merge_llm_findings(all_findings, llm_findings)

    regression_findings = _detect_regressions_from_diff(diff_base)

    if not summary_only:
        changed_ranges = get_changed_line_ranges(diff_base)
        findings = filter_findings_to_diff(all_findings, changed_ranges)
        # Regression findings ARE the diff — no need to filter them
        findings.extend(regression_findings)
    else:
        findings = all_findings + regression_findings

    all_findings.extend(regression_findings)

    if findings and not summary_only:
        _post_pr_review(findings[:max_comments], pr_number, repo)

    _post_summary_comment(
        all_findings,
        findings,
        pr_number,
        repo,
        grade=grade,
        previous_grade=previous_grade,
    )

    console.print(
        f"[green]Posted review on PR #{pr_number} "
        f"({len(findings)} inline, {len(all_findings)} total)[/green]"
    )


def get_changed_line_ranges(base_ref: str = "origin/main") -> list[dict]:
    try:
        result = subprocess.run(
            ["git", "diff", "--unified=0", f"{base_ref}...HEAD"],
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            return []
    except FileNotFoundError:
        return []

    return _parse_unified_diff(result.stdout)


def _parse_unified_diff(diff_output: str) -> list[dict]:
    entries = []
    current_file = None

    for line in diff_output.splitlines():
        if line.startswith("+++ b/"):
            current_file = line[6:]
            continue

        hunk_match = re.match(r"^@@ .+ \+(\d+)(?:,(\d+))? @@", line)
        if hunk_match and current_file:
            start = int(hunk_match.group(1))
            count = int(hunk_match.group(2) or 1)
            if count > 0:
                entries.append(
                    {
                        "file": current_file,
                        "start": start,
                        "end": start + count - 1,
                    }
                )

    return entries


def _get_per_file_diffs(base_ref: str = "origin/main") -> dict[str, str]:
    """Return a dict mapping file paths to their individual diff text."""
    try:
        result = subprocess.run(
            ["git", "diff", "--unified=3", f"{base_ref}...HEAD"],
            capture_output=True,
            text=True,
        )
        if result.returncode != 0:
            return {}
    except FileNotFoundError:
        return {}

    file_diffs: dict[str, str] = {}
    current_file = None
    current_lines: list[str] = []

    for line in result.stdout.splitlines():
        if line.startswith("diff --git"):
            if current_file and current_lines:
                file_diffs[current_file] = "\n".join(current_lines)
            current_file = None
            current_lines = [line]
        elif line.startswith("+++ b/"):
            current_file = line[6:]
            current_lines.append(line)
        else:
            current_lines.append(line)

    if current_file and current_lines:
        file_diffs[current_file] = "\n".join(current_lines)

    return file_diffs


def _detect_regressions_from_diff(base_ref: str = "origin/main") -> list[dict]:
    """Run security regression detection on the PR diff."""
    file_diffs = _get_per_file_diffs(base_ref)
    regression_findings: list[dict] = []

    for file_path, diff_text in file_diffs.items():
        findings = detect_security_regressions(diff_text, file_path)
        for f in findings:
            regression_findings.append(
                {
                    "file": f.get("file", ""),
                    "line": f.get("line", 1),
                    "message": f.get("message", ""),
                    "rule_id": f.get("rule_id", ""),
                    "severity": f.get("severity", "HIGH"),
                    "category": "security_regression",
                    "control_type": f.get("control_type", ""),
                    "kind": "security_regression",
                }
            )

    return regression_findings


def filter_findings_to_diff(
    findings: list[dict], changed_ranges: list[dict]
) -> list[dict]:
    if not changed_ranges:
        return []

    ranges_by_file = {}
    for r in changed_ranges:
        ranges_by_file.setdefault(r["file"], []).append((r["start"], r["end"]))

    filtered = []
    for finding in findings:
        file = finding.get("file", "")
        line = finding.get("line", 0)

        file_ranges = ranges_by_file.get(file, [])
        if not file_ranges:
            for diff_file, ranges in ranges_by_file.items():
                if file.endswith("/" + diff_file) or diff_file.endswith("/" + file):
                    file_ranges = ranges
                    break
        for start, end in file_ranges:
            if start <= line <= end:
                filtered.append(finding)
                break

    return filtered


def _flatten_findings(results: dict) -> list[dict]:
    findings = []

    for category in ("danger", "quality", "secrets", "custom_rules"):
        for f in results.get(category, []) or []:
            findings.append(
                {
                    "file": f.get("file") or f.get("file_path") or "",
                    "line": f.get("line") or f.get("line_number") or 1,
                    "message": f.get("message")
                    or f.get("msg")
                    or f.get("detail")
                    or "",
                    "rule_id": f.get("rule_id") or "",
                    "severity": f.get("severity", "MEDIUM"),
                    "category": category,
                }
            )

    return findings


def _merge_llm_findings(
    static_findings: list[dict], llm_findings: list[dict]
) -> list[dict]:
    llm_by_loc: dict[tuple, dict] = {}
    for f in llm_findings:
        file = f.get("file", "")
        line = f.get("line", 0)
        key = (os.path.basename(file), line)
        llm_by_loc[key] = f

    matched_keys = set()
    for finding in static_findings:
        file = finding.get("file", "")
        line = finding.get("line", 0)
        key = (os.path.basename(file), line)
        if key in llm_by_loc:
            llm = llm_by_loc[key]
            if llm.get("suggestion"):
                finding["suggestion"] = llm["suggestion"]
            if llm.get("explanation"):
                finding["explanation"] = llm["explanation"]
            if llm.get("vulnerable_code"):
                finding["vulnerable_code"] = llm["vulnerable_code"]
            if llm.get("fixed_code"):
                finding["fixed_code"] = llm["fixed_code"]
            matched_keys.add(key)

    for key, llm in llm_by_loc.items():
        if key not in matched_keys:
            static_findings.append(
                {
                    "file": llm.get("file", ""),
                    "line": llm.get("line", 0),
                    "message": llm.get("message", ""),
                    "rule_id": llm.get("rule_id", ""),
                    "severity": llm.get("severity", "MEDIUM"),
                    "category": llm.get("_category", "security"),
                    "suggestion": llm.get("suggestion"),
                    "explanation": llm.get("explanation"),
                    "vulnerable_code": llm.get("vulnerable_code"),
                    "fixed_code": llm.get("fixed_code"),
                    "_source": "llm",
                }
            )

    return static_findings


_REGRESSION_SUGGESTIONS: dict[str, str] = {
    "auth": "Re-add the authentication decorator or dependency. Removing auth exposes the endpoint to unauthenticated access.",
    "csrf": "Re-enable CSRF protection. Without it, the endpoint is vulnerable to cross-site request forgery attacks.",
    "tls": "Re-enable TLS certificate verification (verify=True). Disabling it allows man-in-the-middle attacks.",
    "crypto": "Use a strong hash algorithm (SHA-256 or better). Weak hashes like MD5/SHA-1 are vulnerable to collision attacks.",
    "rate_limit": "Re-add rate limiting. Without it, the endpoint is vulnerable to brute-force and denial-of-service attacks.",
    "validation": "Re-add input validation or sanitization. Without it, the endpoint may be vulnerable to injection attacks.",
    "headers": "Re-add the security header or middleware. Security headers protect against XSS, clickjacking, and other attacks.",
    "encryption": "Re-add encryption. Removing it may expose sensitive data in plaintext.",
    "logging": "Re-add audit logging. Without it, security-relevant actions go untracked.",
    "sanitization": "Re-add output sanitization. Without it, user-supplied content may cause XSS or injection vulnerabilities.",
    "permission": "Re-add the permission check. Removing it may allow unauthorized access to restricted resources.",
}

_RULE_SUGGESTIONS: dict[str, str] = {
    "SKY-D201": "Replace `eval()` with `json.loads()`, `ast.literal_eval()`, or a safe parser. Never evaluate untrusted input.",
    "SKY-D203": "Replace `os.system()` with `subprocess.run()` with `shell=False`. Pass arguments as a list.",
    "SKY-D211": "Use parameterized queries: `cursor.execute('SELECT * FROM t WHERE x = ?', (val,))` instead of f-strings.",
    "SKY-D212": "Sanitize input before passing to shell commands, or use `subprocess.run()` with `shell=False` and argument lists.",
    "SKY-D215": "Validate file paths against an allowed directory: `Path(path).resolve().relative_to(allowed_dir)`.",
    "SKY-D216": "Validate URLs against an allowlist of domains. Block internal IPs (`127.0.0.1`, `169.254.x.x`, `10.x.x.x`).",
    "SKY-D223": "Add the package to `requirements.txt` or `pyproject.toml`, or remove the import if unused.",
    "SKY-S101": "Move secrets to environment variables: `os.getenv('SECRET_KEY')`. Never hardcode credentials.",
}


def _format_review_comment(finding: dict) -> str:
    kind = finding.get("kind", "")
    severity = finding.get("severity", "MEDIUM")
    rule_id = finding.get("rule_id", "")
    message = finding.get("message", "")
    rule_str = f" `{rule_id}`" if rule_id else ""

    if kind == "security_regression":
        control_type = finding.get("control_type", "")
        control_label = control_type.replace("_", " ").title() if control_type else "Unknown"
        parts = [
            f"⚠️ **SECURITY REGRESSION**{rule_str} — {control_label}",
            "",
            message,
        ]
        suggestion = _REGRESSION_SUGGESTIONS.get(control_type)
        if suggestion:
            parts.extend(["", f"**Fix:** {suggestion}"])
    else:
        badge = {"CRITICAL": "🔴", "HIGH": "🟠", "MEDIUM": "🟡", "LOW": "🔵"}.get(
            severity, "⚪"
        )
        parts = [f"{badge} **{severity}**{rule_str}", "", message]

        explanation = finding.get("explanation")
        if explanation:
            parts.extend(["", f"**Why:** {explanation}"])

        vulnerable_code = finding.get("vulnerable_code")
        fixed_code = finding.get("fixed_code")

        if vulnerable_code and fixed_code:
            parts.extend(
                [
                    "",
                    "**Vulnerable code:**",
                    "```python",
                    vulnerable_code,
                    "```",
                    "",
                    "**Fixed code:**",
                    "```python",
                    fixed_code,
                    "```",
                ]
            )
        else:
            suggestion = finding.get("suggestion") or _RULE_SUGGESTIONS.get(rule_id)
            if suggestion:
                parts.extend(["", f"**Fix:** {suggestion}"])

    footer = "\n\n---\n_🤖 Analyzed by [Skylos](https://github.com/duriantaco/skylos) • [Add to your repo](https://github.com/duriantaco/skylos#cicd)_"
    parts.append(footer)

    return "\n".join(parts)


def _to_relative_path(filepath: str) -> str:
    try:
        result = subprocess.run(
            ["git", "rev-parse", "--show-toplevel"],
            capture_output=True,
            text=True,
        )
        if result.returncode == 0:
            root = result.stdout.strip()
            if filepath.startswith(root):
                return filepath[len(root) :].lstrip("/")
    except Exception:
        pass
    return filepath


def _post_pr_review(findings: list[dict], pr_number: int, repo: str) -> None:
    comments = []
    for f in findings:
        if not f.get("file") or not f.get("line"):
            continue
        comments.append(
            {
                "path": _to_relative_path(f["file"]),
                "line": f["line"],
                "body": _format_review_comment(f),
            }
        )

    if not comments:
        return

    payload = {
        "body": (
            f"Skylos found {len(comments)} issue(s) on changed lines.\n\n"
            "---\n"
            "_🤖 Analyzed by [Skylos](https://github.com/duriantaco/skylos) • "
            "[Set up in 30 seconds](https://github.com/duriantaco/skylos#cicd)_"
        ),
        "event": "COMMENT",
        "comments": comments,
    }

    try:
        subprocess.run(
            [
                "gh",
                "api",
                "--method",
                "POST",
                f"/repos/{repo}/pulls/{pr_number}/reviews",
                "--input",
                "-",
            ],
            input=json.dumps(payload),
            capture_output=True,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        console.print(f"[yellow]Failed to post PR review: {e.stderr}[/yellow]")


def _post_summary_comment(
    all_findings: list[dict],
    diff_findings: list[dict],
    pr_number: int,
    repo: str,
    *,
    grade: dict | None = None,
    previous_grade: dict | None = None,
) -> None:
    by_severity = {}
    for f in all_findings:
        sev = f.get("severity", "MEDIUM")
        by_severity[sev] = by_severity.get(sev, 0) + 1

    by_category = {}
    for f in all_findings:
        cat = f.get("category", "other")
        by_category[cat] = by_category.get(cat, 0) + 1

    lines = [
        "## Skylos Analysis Summary",
        "",
        f"**{len(diff_findings)}** issue(s) on changed lines | "
        f"**{len(all_findings)}** total",
        "",
        "| Severity | Count |",
        "|----------|-------|",
    ]

    for sev in ("CRITICAL", "HIGH", "MEDIUM", "LOW"):
        count = by_severity.get(sev, 0)
        if count > 0:
            lines.append(f"| {sev} | {count} |")

    if by_category:
        lines.extend(
            [
                "",
                "| Category | Count |",
                "|----------|-------|",
            ]
        )
        for cat in ("danger", "quality", "secrets", "custom_rules", "security_regression"):
            count = by_category.get(cat, 0)
            if count > 0:
                lines.append(f"| {cat} | {count} |")

    regression_findings = [
        f for f in diff_findings if f.get("kind") == "security_regression"
    ]
    if regression_findings:
        lines.extend(
            [
                "",
                "### ⚠️ Security Regressions Detected",
                "",
                "| Control | File | Message |",
                "|---------|------|---------|",
            ]
        )
        for f in regression_findings:
            control = f.get("control_type", "unknown")
            file = os.path.basename(f.get("file", ""))
            msg = f.get("message", "")
            lines.append(f"| {control} | {file} | {msg} |")

    critical_findings = [
        f for f in diff_findings if f.get("severity") in ("CRITICAL", "HIGH")
    ]
    if critical_findings:
        lines.extend(["", "### Top Issues", ""])
        for f in critical_findings[:5]:
            sev = f.get("severity", "MEDIUM")
            badge = {"CRITICAL": "🔴", "HIGH": "🟠"}.get(sev, "🟡")
            rule = f" `{f['rule_id']}`" if f.get("rule_id") else ""
            file = os.path.basename(f.get("file", ""))
            line_no = f.get("line", "")
            loc = f" ({file}:{line_no})" if file else ""
            lines.append(f"- {badge} **{sev}**{rule}{loc}: {f.get('message', '')}")

            vuln_code = f.get("vulnerable_code")
            fix_code = f.get("fixed_code")
            if vuln_code and fix_code:
                lines.append("")
                lines.append("  <details><summary>View fix</summary>")
                lines.append("")
                lines.append("  **Vulnerable:**")
                lines.append("  ```python")
                for code_line in vuln_code.splitlines():
                    lines.append(f"  {code_line}")
                lines.append("  ```")
                lines.append("  **Fixed:**")
                lines.append("  ```python")
                for code_line in fix_code.splitlines():
                    lines.append(f"  {code_line}")
                lines.append("  ```")
                lines.append("  </details>")
                lines.append("")
            else:
                fix = f.get("suggestion") or _RULE_SUGGESTIONS.get(f.get("rule_id", ""))
                if fix:
                    lines.append(f"  - **Fix:** {fix}")

    if grade:
        overall = grade["overall"]
        cats = grade["categories"]

        lines.extend(["", "### Codebase Grade", ""])

        if previous_grade:
            prev = previous_grade["overall"]
            delta = overall["score"] - prev["score"]
            arrow = "+" if delta > 0 else ""
            direction = "\u2191" if delta > 0 else ("\u2193" if delta < 0 else "\u2194")
            lines.append(
                f"**{prev['letter']} ({prev['score']}) \u2192 "
                f"{overall['letter']} ({overall['score']}) {direction}** "
                f"({arrow}{delta})"
            )
        else:
            lines.append(f"**Overall: {overall['letter']} ({overall['score']}/100)**")

        lines.extend(
            [
                "",
                "| Category | Score | Grade | Key Issue |",
                "|----------|-------|-------|-----------|",
            ]
        )

        for cat_name in ("security", "quality", "dead_code", "dependencies", "secrets"):
            cat = cats[cat_name]
            display = cat_name.replace("_", " ").title()
            issue = (cat.get("key_issue") or "-")[:50]

            delta_str = ""
            if previous_grade and cat_name in previous_grade.get("categories", {}):
                prev_cat = previous_grade["categories"][cat_name]
                cat_delta = cat["score"] - prev_cat["score"]
                if cat_delta != 0:
                    d_arrow = "\u2191" if cat_delta > 0 else "\u2193"
                    delta_str = f" {d_arrow}{abs(cat_delta)}"

            lines.append(
                f"| {display} | {cat['score']}{delta_str} | {cat['letter']} | {issue} |"
            )

    body = "\n".join(lines)

    try:
        subprocess.run(
            ["gh", "pr", "comment", str(pr_number), "--body", body, "--repo", repo],
            capture_output=True,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        console.print(f"[yellow]Failed to post summary comment: {e.stderr}[/yellow]")


def _fetch_previous_grade(repo: str, base_branch: str = "origin/main") -> dict | None:
    try:
        from skylos.api import get_project_token, BASE_URL
        import requests

        token = get_project_token()
        if not token:
            return None

        branch = base_branch.replace("origin/", "")
        resp = requests.get(
            f"{BASE_URL}/api/grade/latest",
            params={"branch": branch},
            headers={"Authorization": f"Bearer {token}"},
            timeout=10,
        )
        if resp.status_code == 200:
            return resp.json().get("grade")
    except Exception:
        pass
    return None


def _detect_pr_number() -> int | None:
    ref = os.environ.get("GITHUB_REF", "")
    match = re.match(r"refs/pull/(\d+)/merge", ref)
    if match:
        return int(match.group(1))
    return None


def _gh_available() -> bool:
    try:
        subprocess.run(["gh", "--version"], capture_output=True, check=True)
        return True
    except (FileNotFoundError, subprocess.CalledProcessError):
        return False



"""Tests for Next.js-specific security pattern detection in TypeScript."""

from __future__ import annotations

import pytest

from skylos.visitors.languages.typescript.danger import (
    _check_nextjs_missing_auth,
    _check_nextjs_client_secrets,
    _check_nextjs_server_action_sqli,
)


# ---------- Missing auth in API routes (SKY-D280) ----------


class TestMissingAuth:
    def test_route_with_post_no_auth(self):
        source = b"""
export async function POST(request: Request) {
    const data = await request.json();
    return Response.json({ ok: true });
}
"""
        findings = []
        _check_nextjs_missing_auth(
            source, "/project/app/api/users/route.ts", findings
        )
        assert len(findings) == 1
        assert findings[0]["rule_id"] == "SKY-D280"

    def test_route_with_post_has_auth(self):
        source = b"""
import { getServerSession } from "next-auth";

export async function POST(request: Request) {
    const session = await getServerSession();
    if (!session) return Response.json({ error: "Unauthorized" }, { status: 401 });
    return Response.json({ ok: true });
}
"""
        findings = []
        _check_nextjs_missing_auth(
            source, "/project/app/api/users/route.ts", findings
        )
        assert len(findings) == 0

    def test_route_with_get_only_no_auth_ok(self):
        """GET-only routes don't need auth (read-only)."""
        source = b"""
export async function GET(request: Request) {
    return Response.json({ items: [] });
}
"""
        findings = []
        _check_nextjs_missing_auth(
            source, "/project/app/api/items/route.ts", findings
        )
        assert len(findings) == 0

    def test_route_with_delete_no_auth(self):
        source = b"""
export async function DELETE(request: Request) {
    await db.delete(items);
    return Response.json({ ok: true });
}
"""
        findings = []
        _check_nextjs_missing_auth(
            source, "/project/app/api/items/route.ts", findings
        )
        assert len(findings) == 1
        assert findings[0]["rule_id"] == "SKY-D280"

    def test_non_route_file_ignored(self):
        source = b"""
export async function POST(data: any) {
    return data;
}
"""
        findings = []
        _check_nextjs_missing_auth(
            source, "/project/src/utils/helpers.ts", findings
        )
        assert len(findings) == 0

    def test_pages_api_route(self):
        source = b"""
export default function handler(req, res) {
    if (req.method === 'POST') {
        res.json({ ok: true });
    }
}

export async function POST(request: Request) {
    return Response.json({});
}
"""
        findings = []
        _check_nextjs_missing_auth(
            source, "/project/pages/api/users.ts", findings
        )
        assert len(findings) == 1

    def test_route_with_cookies_auth(self):
        source = b"""
import { cookies } from "next/headers";

export async function POST(request: Request) {
    const cookieStore = cookies();
    return Response.json({ ok: true });
}
"""
        findings = []
        _check_nextjs_missing_auth(
            source, "/project/app/api/data/route.ts", findings
        )
        assert len(findings) == 0


# ---------- Client component secrets (SKY-S102) ----------


class TestClientSecrets:
    def test_server_env_in_client_component(self):
        source = b"""
"use client"

export default function Dashboard() {
    const apiKey = process.env.DATABASE_URL;
    return <div>Dashboard</div>;
}
"""
        findings = []
        _check_nextjs_client_secrets(source, "/project/app/dashboard/page.tsx", findings)
        assert len(findings) == 1
        assert findings[0]["rule_id"] == "SKY-S102"
        assert "DATABASE_URL" in findings[0]["message"]

    def test_next_public_env_in_client_ok(self):
        source = b"""
"use client"

export default function Dashboard() {
    const url = process.env.NEXT_PUBLIC_API_URL;
    return <div>Dashboard</div>;
}
"""
        findings = []
        _check_nextjs_client_secrets(source, "/project/app/dashboard/page.tsx", findings)
        assert len(findings) == 0

    def test_server_component_env_ok(self):
        """Server components can access any env var."""
        source = b"""
export default function Dashboard() {
    const dbUrl = process.env.DATABASE_URL;
    return <div>Dashboard</div>;
}
"""
        findings = []
        _check_nextjs_client_secrets(source, "/project/app/dashboard/page.tsx", findings)
        assert len(findings) == 0

    def test_multiple_secret_envs(self):
        source = b"""
"use client"

const db = process.env.DATABASE_URL;
const secret = process.env.JWT_SECRET;
const pub = process.env.NEXT_PUBLIC_NAME;
"""
        findings = []
        _check_nextjs_client_secrets(source, "/project/app/component.tsx", findings)
        assert len(findings) == 2
        messages = [f["message"] for f in findings]
        assert any("DATABASE_URL" in m for m in messages)
        assert any("JWT_SECRET" in m for m in messages)

    def test_use_client_with_semicolon(self):
        source = b"""'use client';

const key = process.env.SECRET_KEY;
"""
        findings = []
        _check_nextjs_client_secrets(source, "/project/app/comp.tsx", findings)
        assert len(findings) == 1


# ---------- SQL injection in server actions (SKY-D281) ----------


class TestServerActionSQLi:
    def test_sql_template_in_server_action(self):
        source = b"""
"use server"

export async function deleteUser(userId: string) {
    await db.query(`DELETE FROM users WHERE id = ${userId}`);
}
"""
        findings = []
        _check_nextjs_server_action_sqli(source, "/project/app/actions.ts", findings)
        assert len(findings) == 1
        assert findings[0]["rule_id"] == "SKY-D281"

    def test_parameterized_query_ok(self):
        source = b"""
"use server"

export async function deleteUser(userId: string) {
    await db.query("DELETE FROM users WHERE id = $1", [userId]);
}
"""
        findings = []
        _check_nextjs_server_action_sqli(source, "/project/app/actions.ts", findings)
        assert len(findings) == 0

    def test_execute_with_template(self):
        source = b"""
"use server"

export async function updateUser(id: string, name: string) {
    await conn.execute(`UPDATE users SET name = ${name} WHERE id = ${id}`);
}
"""
        findings = []
        _check_nextjs_server_action_sqli(source, "/project/app/actions.ts", findings)
        assert len(findings) == 1

    def test_non_server_action_ignored(self):
        """Regular files with SQL templates are caught by general SQL injection check."""
        source = b"""
export async function deleteUser(userId: string) {
    await db.query(`DELETE FROM users WHERE id = ${userId}`);
}
"""
        findings = []
        _check_nextjs_server_action_sqli(source, "/project/app/actions.ts", findings)
        assert len(findings) == 0

    def test_raw_method(self):
        source = b"""
"use server"

export async function search(term: string) {
    return await prisma.raw(`SELECT * FROM items WHERE name = ${term}`);
}
"""
        findings = []
        _check_nextjs_server_action_sqli(source, "/project/app/actions.ts", findings)
        assert len(findings) == 1

    def test_template_without_sql_keywords_ok(self):
        source = b"""
"use server"

export async function doThing(name: string) {
    await db.query(`hello ${name}`);
}
"""
        findings = []
        _check_nextjs_server_action_sqli(source, "/project/app/actions.ts", findings)
        assert len(findings) == 0



from __future__ import annotations

import hashlib
import subprocess
import sys
import shutil
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path


@dataclass
class TestResult:
    passed: bool
    output: str = ""
    command: str = ""
    duration_s: float = 0.0


@dataclass
class VerifyResult:
    finding_resolved: bool
    remaining_rule_ids: list[str] = field(default_factory=list)


class RemediationExecutor:
    def __init__(self, *, test_cmd: str | None = None, project_root: str | Path = "."):
        self.test_cmd = test_cmd
        self.project_root = Path(project_root).resolve()
        self._backups: dict[str, str] = {}

    def apply_fix(self, file_path: str, fixed_code: str) -> bool:
        p = Path(file_path)
        if not p.exists():
            return False
        try:
            original = p.read_text(encoding="utf-8")
            self._backups[str(p)] = original
            p.write_text(fixed_code, encoding="utf-8")
            return True
        except OSError:
            return False

    def revert_fix(self, file_path: str) -> bool:
        key = str(Path(file_path))
        original = self._backups.pop(key, None)
        if original is None:
            return False
        try:
            Path(file_path).write_text(original, encoding="utf-8")
            return True
        except OSError:
            return False

    def revert_all(self):
        for fp in list(self._backups.keys()):
            self.revert_fix(fp)

    def run_tests(self, timeout: int = 300) -> TestResult:
        cmd = self.test_cmd or self._detect_test_command()
        if not cmd:
            return TestResult(passed=True, output="No test suite detected.", command="")

        import time

        start = time.monotonic()
        try:
            result = subprocess.run(
                cmd,
                shell=True,
                capture_output=True,
                text=True,
                timeout=timeout,
                cwd=str(self.project_root),
            )
            duration = time.monotonic() - start
            return TestResult(
                passed=result.returncode == 0,
                output=(result.stdout + result.stderr)[-5000:],
                command=cmd,
                duration_s=round(duration, 2),
            )
        except subprocess.TimeoutExpired:
            duration = time.monotonic() - start
            return TestResult(
                passed=False,
                output=f"Test suite timed out after {timeout}s",
                command=cmd,
                duration_s=round(duration, 2),
            )
        except OSError as e:
            return TestResult(passed=False, output=str(e), command=cmd)

    def _detect_test_command(self) -> str | None:
        root = self.project_root

        pyproject = root / "pyproject.toml"
        if pyproject.exists():
            content = pyproject.read_text(encoding="utf-8", errors="ignore")
            if "[tool.pytest" in content or "pytest" in content:
                return "python -m pytest --tb=short -q"

        if (root / "pytest.ini").exists():
            return "python -m pytest --tb=short -q"
        if (root / "setup.cfg").exists():
            cfg = (root / "setup.cfg").read_text(encoding="utf-8", errors="ignore")
            if "[tool:pytest]" in cfg:
                return "python -m pytest --tb=short -q"
        if (root / "tox.ini").exists():
            return "tox -e py --quiet"

        makefile = root / "Makefile"
        if makefile.exists():
            content = makefile.read_text(encoding="utf-8", errors="ignore")
            if "test:" in content:
                return "make test"

        if (root / "test").is_dir() or (root / "tests").is_dir():
            return "python -m pytest --tb=short -q"

        return None

    def verify_fix(self, file_path: str, original_rule_ids: list[str]) -> VerifyResult:
        import json as _json
        from skylos.analyzer import analyze as run_analyze

        try:
            raw = run_analyze(
                file_path,
                conf=0,
                enable_danger=True,
                enable_quality=True,
                enable_secrets=True,
            )
            result = _json.loads(raw) if isinstance(raw, str) else raw
        except Exception:
            return VerifyResult(finding_resolved=True)

        remaining = set()
        for key in ("danger", "quality", "secrets"):
            for finding in result.get(key, []) or []:
                rid = finding.get("rule_id", "")
                if rid in original_rule_ids:
                    remaining.add(rid)

        return VerifyResult(
            finding_resolved=len(remaining) == 0,
            remaining_rule_ids=sorted(remaining),
        )

    def create_branch(self, prefix: str = "skylos/fix") -> str:
        ts = datetime.now().strftime("%Y%m%d-%H%M%S")
        short = hashlib.sha1(ts.encode()).hexdigest()[:6]
        branch = f"{prefix}-{ts[:8]}-{short}"
        subprocess.run(
            ["git", "checkout", "-b", branch],
            cwd=str(self.project_root),
            capture_output=True,
            check=True,
        )
        return branch

    def commit_fixes(self, message: str, files: list[str]) -> bool:
        try:
            for f in files:
                subprocess.run(
                    ["git", "add", f],
                    cwd=str(self.project_root),
                    capture_output=True,
                    check=True,
                )
            subprocess.run(
                ["git", "commit", "-m", message],
                cwd=str(self.project_root),
                capture_output=True,
                check=True,
            )
            return True
        except subprocess.CalledProcessError:
            return False

    def push_branch(self, branch: str) -> bool:
        try:
            subprocess.run(
                ["git", "push", "-u", "origin", branch],
                cwd=str(self.project_root),
                capture_output=True,
                check=True,
            )
            return True
        except subprocess.CalledProcessError:
            return False

    def create_pr(
        self, branch: str, title: str, body: str, base: str = "main"
    ) -> str | None:
        if not shutil.which("gh"):
            print("Warning: gh CLI not found, skipping PR creation.", file=sys.stderr)
            return None
        try:
            result = subprocess.run(
                [
                    "gh",
                    "pr",
                    "create",
                    "--title",
                    title,
                    "--body",
                    body,
                    "--base",
                    base,
                    "--head",
                    branch,
                ],
                cwd=str(self.project_root),
                capture_output=True,
                text=True,
                check=True,
            )
            return result.stdout.strip()
        except subprocess.CalledProcessError as e:
            print(f"PR creation failed: {e.stderr}", file=sys.stderr)
            return None



from pathlib import Path
from skylos.engines.go_runner import run_go_engine_for_module
from skylos.visitor import Definition
from skylos.visitors.languages.go.quality import scan_go_quality, GO_LANG

try:
    from tree_sitter import Parser
except ImportError:
    Parser = None


class _GoDummyVisitor:
    def __init__(self):
        self.is_test_file = False
        self.test_decorated_lines = set()
        self.dataclass_fields = set()
        self.pydantic_models = set()
        self.class_defs = {}
        self.first_read_lineno = {}
        self.framework_decorated_lines = set()


_GO_RULE_REMAP = {
    "SKY-G211": "SKY-D211",  # SQL injection
    "SKY-G212": "SKY-D212",  # Command injection
    "SKY-G215": "SKY-D215",  # Path traversal
    "SKY-G216": "SKY-D216",  # SSRF
    "SKY-G207": "SKY-D207",  # Weak hash MD5
    "SKY-G208": "SKY-D208",  # Weak hash SHA1
    "SKY-G210": "SKY-D210",  # TLS verification disabled
    "SKY-G220": "SKY-D230",  # Open redirect
}

_go_module_cache = {}


def clear_go_cache():
    _go_module_cache.clear()


def _get_module_result(module_root):
    key = str(module_root)
    if key not in _go_module_cache:
        try:
            _go_module_cache[key] = run_go_engine_for_module(module_root)
        except Exception as e:
            import os

            if os.getenv("SKYLOS_DEBUG"):
                print(f"Go analysis failed: {e}")
            _go_module_cache[key] = {"findings": [], "symbols": None}
    return _go_module_cache[key]


def _convert_symbols(symbols_data, file_path):
    if not symbols_data:
        return [], []

    file_str = str(file_path)
    defs = []
    refs = []

    for d in symbols_data.get("defs", []):
        if d.get("file") != file_str:
            continue
        defn = Definition(
            name=d["name"],
            t=d["type"],
            filename=file_path,
            line=d.get("line", 0),
        )
        defn.is_exported = d.get("is_exported", False)
        if defn.is_exported:
            defn.references = 1
        defs.append(defn)

    for r in symbols_data.get("refs", []):
        if r.get("file") == file_str:
            refs.append((r["name"], r["file"]))

    return defs, refs


def scan_go_file(file_path, cfg):
    file_path = Path(file_path)

    module_root = _find_module_root(file_path)

    if not module_root:
        module_root = file_path.parent

    result = _get_module_result(module_root)

    findings = result.get("findings", [])
    file_findings = [
        f for f in findings if Path(f.get("file", "")).resolve() == file_path.resolve()
    ]

    for f in file_findings:
        rid = f.get("rule_id", "")
        if rid in _GO_RULE_REMAP:
            f["rule_id"] = _GO_RULE_REMAP[rid]

    symbols_data = result.get("symbols")
    defs, refs = _convert_symbols(symbols_data, str(file_path.resolve()))

    # Run tree-sitter-based quality checks
    quality_findings = []
    if Parser is not None and GO_LANG is not None:
        try:
            source = file_path.read_bytes()
            parser = Parser(GO_LANG)
            tree = parser.parse(source)
            quality_findings = scan_go_quality(
                tree.root_node,
                source,
                str(file_path),
            )
        except Exception:
            import os

            if os.getenv("SKYLOS_DEBUG"):
                import traceback

                print(
                    f"Go quality scan failed for {file_path}: {traceback.format_exc()}"
                )

    return (
        defs,  # 0: definitions
        refs,  # 1: references
        set(),  # 2: dynamic refs
        set(),  # 3: exports
        _GoDummyVisitor(),  # 4: test_flags
        _GoDummyVisitor(),  # 5: framework_flags
        quality_findings,  # 6: quality findings
        file_findings,  # 7: danger/security findings
        [],  # 8: pro_finds
        None,  # 9: pattern_tracker
        None,  # 10: empty_file_finding
        cfg,  # 11: config
        [],  # 12: raw_imports
    )


def _find_module_root(file_path):
    current = Path(file_path).parent
    while current != current.parent:
        if (current / "go.mod").exists():
            return current
        current = current.parent
    return None



import ast
from skylos.rules.danger.danger_cors.cors_flow import scan


def _scan_code(code, filename="app.py"):
    tree = ast.parse(code)
    findings = []
    scan(tree, filename, findings)
    return findings


def _rule_ids(findings):
    return {f["rule_id"] for f in findings}


def test_flask_cors_no_origins():
    code = (
        "from flask_cors import CORS\n"
        "from flask import Flask\n"
        "app = Flask(__name__)\n"
        "CORS(app)\n"
    )
    findings = _scan_code(code)
    assert "SKY-D231" in _rule_ids(findings)


def test_django_cors_allow_all():
    code = "CORS_ALLOW_ALL_ORIGINS = True\n"
    findings = _scan_code(code)
    assert "SKY-D231" in _rule_ids(findings)


def test_cors_with_origins_safe():
    code = (
        "from flask_cors import CORS\n"
        "from flask import Flask\n"
        "app = Flask(__name__)\n"
        "CORS(app, origins=['https://example.com'])\n"
    )
    findings = _scan_code(code)
    assert "SKY-D231" not in _rule_ids(findings)


def test_django_cors_allow_all_false_safe():
    code = "CORS_ALLOW_ALL_ORIGINS = False\n"
    findings = _scan_code(code)
    assert "SKY-D231" not in _rule_ids(findings)
